import sys


if sys.version_info < (3, 8):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution


def _find_pex3_entrypoint():
    pex_dist = distribution("pex")
    for entrypoint in pex_dist.entry_points:
        if entrypoint.name == "pex3":
            return entrypoint
    msg = "Unable to find entrypoint for `pex3`"
    raise ValueError(msg)


if __name__ == "__main__":
    pex3 = _find_pex3_entrypoint()
    pex3_cli = pex3.load()
    sys.exit(pex3_cli())



import contextlib
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import Tuple

import attrs
import pandas
import pyarrow
import pyarrow.compute as pc
import pyarrow.dataset

from tecton_core import conf
from tecton_core import errors
from tecton_core import schema
from tecton_core.offline_store import DEFAULT_OPTIONS_PROVIDERS
from tecton_core.offline_store import DeltaReader
from tecton_core.offline_store import JoinKeyBoundaries
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.offline_store import OfflineStoreReaderParams
from tecton_core.offline_store import ParquetReader
from tecton_core.offline_store import get_s3_options_for_fd
from tecton_core.query.dialect import Dialect
from tecton_core.query.duckdb.rewrite import DuckDBTreeRewriter
from tecton_core.query.errors import UserDefinedTransformationError
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.executor_utils import DebugOutput
from tecton_core.query.executor_utils import QueryTreeMonitor
from tecton_core.query.executor_utils import get_stage_type_for_dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import QueryNode
from tecton_core.query.node_interface import recurse_query_tree
from tecton_core.query.node_utils import get_data_source_dialect
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_core.query.node_utils import get_pipeline_dialect
from tecton_core.query.node_utils import get_staging_nodes
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import MockDataSourceScanNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import StagedTableScanNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import TextEmbeddingInferenceNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.pandas.rewrite import PandasTreeRewriter
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import ModelInferenceCompute
from tecton_core.query.query_tree_compute import QueryTreeCompute
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.secrets import SecretResolver
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage


logger = logging.getLogger(__name__)


def _pyarrow_type_contains_map_type(pyarrow_type: pyarrow.DataType) -> bool:
    if isinstance(pyarrow_type, pyarrow.MapType):
        return True
    elif isinstance(pyarrow_type, pyarrow.StructType):
        return any(_pyarrow_type_contains_map_type(field.type) for field in pyarrow_type)
    elif isinstance(pyarrow_type, pyarrow.ListType):
        return _pyarrow_type_contains_map_type(pyarrow_type.value_type)
    return False


@dataclass
class QueryTreeOutput:
    # Maps name to pyarrow batch reader containing a data source
    data_source_readers: Dict[str, pyarrow.RecordBatchReader]
    # Maps name to pyarrow batch reader containing a feature view
    feature_view_readers: Dict[str, pyarrow.RecordBatchReader]
    odfv_input: Optional[pyarrow.RecordBatchReader] = None
    odfv_output: Optional[pyarrow.RecordBatchReader] = None

    @property
    def result_df(self) -> pandas.DataFrame:
        if self.odfv_output is not None:
            return self.odfv_output.read_pandas()
        assert self.odfv_input is not None

        contains_map_type = any(_pyarrow_type_contains_map_type(field.type) for field in self.odfv_input.schema)
        if contains_map_type:
            # The `maps_as_pydicts` parameter for pyarrow.Table.to_pandas is only supported starting in pyarrow 13.0.0.
            if pyarrow.__version__ < "13.0.0":
                msg = f"Rift requires pyarrow>=13.0.0 to perform feature retrieval for Map features. You have version {pyarrow.__version__}."
                raise RuntimeError(msg)
            return self.odfv_input.read_pandas(maps_as_pydicts="strict")

        return self.odfv_input.read_pandas()

    @property
    def result_table(self) -> pyarrow.Table:
        assert self.odfv_output is None, "Can't retrieve ODFV output as a pyarrow table"
        return self.odfv_input


def checkpoint(
    root: NodeRef, query_tree_step: Optional[QueryTreeStep] = None, staging_node: Optional[QueryNode] = None
) -> None:
    """Checkpointing in QueryTreeExecutor is somewhat similar to one in the Spark DAG:
    (1) QT is being executed up until StagingNode
    (2) The result of this execution is then cached (as pyarrow Table / RecordBatchReader)
    (3) Corresponding StagingNode is replaced with StagedTableScanNode to read cached result on the next stage

    However, for stages that return RecordBatchReader the result is lazily evaluated and will be
    completely unrolled by the next stage.
    """

    def traverse(tree: NodeRef) -> None:
        for i in tree.inputs:
            traverse(tree=i)

        if isinstance(tree.node, StagingNode) and (
            (staging_node and tree.node == staging_node)
            or (query_tree_step and query_tree_step == tree.node.query_tree_step)
        ):
            tree.node = StagedTableScanNode.from_staging_node(tree.node.dialect, tree.node.compute_mode, tree.node)

    traverse(root)


def rewrite_offline_scan_node(root: NodeRef, callback: Callable[[OfflineStoreScanNode, str], None]) -> None:
    def traverse(tree: NodeRef) -> None:
        for i in tree.inputs:
            traverse(tree=i)

        if isinstance(tree.node, OfflineStoreScanNode):
            fdw = tree.node.feature_definition_wrapper
            staged_table_name = f"{fdw.name}_offline_store_scan_{tree.node.node_id.hex[:16]}"

            callback(tree.node, staged_table_name)

            tree.node = StagedTableScanNode(
                tree.node.dialect,
                tree.node.compute_mode,
                staged_columns=tree.node.columns,
                staging_table_name=staged_table_name,
            )

    traverse(root)


UserErrors = (UserDefinedTransformationError,)


@attrs.define
class QueryTreeExecutor:
    offline_store_options_providers: Iterable[OfflineStoreOptionsProvider] = DEFAULT_OPTIONS_PROVIDERS
    secret_resolver: Optional[SecretResolver] = None
    monitor: QueryTreeMonitor = DebugOutput()
    is_debug: bool = attrs.field(init=False)
    # Used to track temp tables per dialect so we can clean them up appropriately & avoid re-registering duplicates
    _dialect_to_temp_table_name: Optional[Dict[Dialect, set]] = attrs.field(init=False)

    def __attrs_post_init__(self):
        # TODO(danny): Expose as configs
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")
        self._dialect_to_temp_table_name = None

    @contextlib.contextmanager
    def _monitor_stage(
        self,
        step: str,
        type_: Optional[int] = None,
        dialect: Optional[Dialect] = None,
    ) -> Iterator[ComputeMonitor]:
        assert type_ or dialect, "Either type or dialect must be provided"
        if not type_:
            type_ = get_stage_type_for_dialect(dialect)

        monitor_stage_id = self.monitor.create_stage(type_, step)

        try:
            self.monitor.update_progress(monitor_stage_id, 0)
            yield ComputeMonitor(
                log_progress=lambda p: self.monitor.update_progress(monitor_stage_id, p),
                set_query=lambda q: self.monitor.set_query(monitor_stage_id, q),
            )
        except UserErrors:
            self.monitor.set_failed(monitor_stage_id, user_error=True)
            raise
        except Exception:
            self.monitor.set_failed(monitor_stage_id, user_error=False)
            raise
        else:
            self.monitor.update_progress(monitor_stage_id, 1)
            self.monitor.set_completed(monitor_stage_id)

    def exec_qt(self, qt_root: NodeRef) -> QueryTreeOutput:
        # Make copy so the execution doesn't mutate the original QT visible to users
        qt_root = qt_root.deepcopy()
        if self.is_debug:
            logger.warning("---------------------------------- Executing overall QT ----------------------------------")
            logger.warning(f"QT: \n{qt_root.pretty_str(columns=True)}")

        rewriter = DuckDBTreeRewriter()
        rewriter.rewrite(qt_root)

        output = self._execute_data_source_step(qt_root)

        # This can only happen if the initial query tree was a single DataSourceScanNode followed by a
        # StagingNode. In that case, we can skip the rest of the query tree.
        if len(output.data_source_readers) == 1 and isinstance(qt_root.node, StagedTableScanNode):
            table_name, pa_reader = output.data_source_readers.popitem()
            return QueryTreeOutput(data_source_readers={}, feature_view_readers={}, odfv_input=pa_reader)

        # Executes the feature view pipeline and stages into memory
        output = self._execute_pipeline_step(output, qt_root)

        if get_first_input_node_of_class(qt_root, OfflineStoreScanNode) is not None:
            with self._monitor_stage("Reading offline store", type_=TectonManagedStage.StageType.OFFLINE_STORE):
                rewrite_offline_scan_node(qt_root, callback=self._process_offline_scan_node)

        output = self._execute_model_inference_step(output, qt_root)

        # Does partial aggregations (if applicable) and spine joins
        qt_root.node = qt_root.node.with_dialect(Dialect.DUCKDB)
        output = self._execute_agg_step(output, qt_root)

        # Runs ODFVs (if applicable)
        output = self._execute_odfv_step(output, qt_root)
        return output

    def _execute_data_source_step(self, qt_root: NodeRef) -> QueryTreeOutput:
        staging_nodes_to_process = get_staging_nodes(qt_root, QueryTreeStep.DATA_SOURCE)
        if len(staging_nodes_to_process) == 0:
            # This is possible if, for example, the querytree is reading from the offline store instead from a data source.
            return QueryTreeOutput(data_source_readers={}, feature_view_readers={})

        data_source_readers = {}
        for name, staging_node in staging_nodes_to_process.items():
            data_source_node_ref = staging_node.input_node
            data_source_node = data_source_node_ref.node

            # No data source dialect will be returned for MockedDataSource
            dialect = get_data_source_dialect(data_source_node_ref) or Dialect.DUCKDB
            with self._monitor_stage("Reading Data Source", dialect=dialect) as monitor:
                compute = SQLCompute.for_dialect(
                    dialect,
                    qt_root,
                    secret_resolver=self.secret_resolver,
                    offline_store_options=self.offline_store_options_providers,
                )

                assert isinstance(data_source_node, (DataSourceScanNode, MockDataSourceScanNode))

                if not isinstance(data_source_node, MockDataSourceScanNode):
                    # TODO: for cases where ds compute != pipeline compute, it maybe better to register a duckdb table
                    compute.register_temp_table_from_data_source(name, data_source_node, self.secret_resolver, monitor)
                    if dialect == get_pipeline_dialect(qt_root) and dialect != Dialect.PANDAS:
                        # No need to export from compute, it will be shared between stages
                        output_pa = None
                    else:
                        expected_schema = (
                            schema.Schema(data_source_node.ds.schema.tecton_schema)
                            if data_source_node.ds.schema
                            else None
                        )
                        output_pa = compute.load_table(table_name=name, expected_output_schema=expected_schema)
                else:
                    # tmp table for user provided data (e.g spine)
                    self._maybe_register_temp_tables(qt_root=qt_root, compute=compute)
                    output_pa = compute.run_sql(data_source_node.to_sql(), return_dataframe=True, monitor=monitor)
                data_source_readers[name] = output_pa

                if output_pa:
                    # checkpoint StagingNode (one at a time) only if there's something to checkpoint
                    checkpoint(qt_root, staging_node=staging_node)

        return QueryTreeOutput(data_source_readers=data_source_readers, feature_view_readers={})

    def _execute_pipeline_step(self, output: QueryTreeOutput, qt_node: NodeRef) -> QueryTreeOutput:
        pipeline_dialect = get_pipeline_dialect(qt_node)
        if not pipeline_dialect:
            # bypass
            return QueryTreeOutput(feature_view_readers=output.data_source_readers, data_source_readers={})

        compute = SQLCompute.for_dialect(pipeline_dialect, qt_node, secret_resolver=self.secret_resolver)
        with self._monitor_stage("Evaluating Feature View pipelines", dialect=pipeline_dialect) as monitor, compute:
            if pipeline_dialect == Dialect.PANDAS:
                rewriter = PandasTreeRewriter()
                rewriter.rewrite(qt_node, compute, output.data_source_readers)
                if self.is_debug:
                    logger.warning(f"PANDAS PRE INIT: \n{qt_node.pretty_str(description=False)}")
            else:
                for table_name, pa_reader in output.data_source_readers.items():
                    if not pa_reader:
                        # assuming that we're sharing computes and table is already loaded
                        continue

                    if self.is_debug:
                        logger.warning(
                            f"Registering staged table {table_name} to pipeline compute with schema:\n{pa_reader.schema}"
                        )
                    compute.register_temp_table(table_name, pa_reader)

            # tmp table for user provided data (e.g spine)
            self._maybe_register_temp_tables(qt_root=qt_node, compute=compute)

            staging_nodes_to_process = get_staging_nodes(qt_node, QueryTreeStep.PIPELINE)
            if len(staging_nodes_to_process) == 0:
                # It's possible that with Pandas transformations there's no staging node,
                # but Pandas rewrite above already replaced Pandas pipeline node with StagedTableScanNode
                return QueryTreeOutput(data_source_readers={}, feature_view_readers={})

            feature_view_tables = self._stage_tables_and_load_pa(
                nodes_to_process=staging_nodes_to_process,
                compute=compute,
                monitor=monitor,
            )
            checkpoint(qt_node, QueryTreeStep.PIPELINE)
            return QueryTreeOutput(data_source_readers={}, feature_view_readers=feature_view_tables)

    def _execute_model_inference_step(self, prev_step_output: QueryTreeOutput, qt_node: NodeRef) -> QueryTreeOutput:
        has_models = get_first_input_node_of_class(qt_node, node_class=TextEmbeddingInferenceNode) is not None

        if not has_models:
            return prev_step_output

        # NOTE: only torch supports model inference
        model_compute = ModelInferenceCompute.for_dialect(Dialect.TORCH)

        with self._monitor_stage("Computing model inference", type_=TectonManagedStage.StageType.PYTHON):
            staging_nodes_to_process = get_staging_nodes(qt_node, QueryTreeStep.MODEL_INFERENCE)
            if len(staging_nodes_to_process) == 0:
                msg = "No `MODEL_INFERENCE` staging nodes, despite having an `TextEmbeddingInferenceNode`"
                raise ValueError(msg)

            # We make a copy since we need to pass through any FVs which do not
            # have a model inference step.
            feature_view_readers = prev_step_output.feature_view_readers.copy()

            for name, staging_node in staging_nodes_to_process.items():
                inference_node_ref = staging_node.input_node
                inference_node = inference_node_ref.node
                assert isinstance(inference_node, TextEmbeddingInferenceNode)

                if not isinstance(inference_node.input_node.node, StagedTableScanNode):
                    msg = "Only supports `StagedTableScanNode` input"
                    raise ValueError(msg)

                input_table_name = inference_node.input_node.node.staging_table_name

                # We get+remove the input_table from feature_view_readers (since we will use `feature_view_readers` to return).
                table_reader = feature_view_readers.pop(input_table_name)
                output_table = model_compute.run_inference(inference_node_ref, table_reader)
                feature_view_readers[staging_node.staging_table_name_unique()] = output_table

            checkpoint(qt_node, QueryTreeStep.MODEL_INFERENCE)
            return QueryTreeOutput(
                data_source_readers={},
                feature_view_readers=feature_view_readers,
            )

    def _execute_agg_step(self, output: QueryTreeOutput, qt_node: NodeRef) -> QueryTreeOutput:
        compute = SQLCompute.for_dialect(Dialect.DUCKDB)

        with self._monitor_stage(
            "Computing aggregated features & joining results", type_=TectonManagedStage.StageType.AGGREGATE
        ) as monitor, compute:
            # The AsOfJoins need access to a spine, which are registered here.
            self._maybe_register_temp_tables(qt_root=qt_node, compute=compute)

            # Register staged pyarrow tables in agg compute
            for table_name, pa_reader in output.feature_view_readers.items():
                if self.is_debug:
                    logger.warning(
                        f"Registering staged table {table_name} to agg compute with schema:\n{pa_reader.schema}"
                    )
                compute.register_temp_table(table_name, pa_reader)

            next_output = self._process_agg_join(output, compute, qt_node, monitor)
            checkpoint(qt_node, QueryTreeStep.AGGREGATION)
            return next_output

    def _execute_odfv_step(self, prev_step_output: QueryTreeOutput, qt_node: NodeRef) -> QueryTreeOutput:
        assert prev_step_output
        assert prev_step_output.odfv_input is not None
        has_odfvs = get_first_input_node_of_class(qt_node, node_class=MultiOdfvPipelineNode) is not None
        if has_odfvs:
            compute = SQLCompute.for_dialect(Dialect.PANDAS)
            # TODO(meastham): Use pyarrow typemapper after upgrading pandas
            with self._monitor_stage(
                "Evaluating On-Demand Feature Views", type_=TectonManagedStage.StageType.PYTHON
            ) as monitor, compute:
                output_reader = compute.run_odfv(qt_node, prev_step_output.odfv_input, monitor=monitor)
        else:
            output_reader = None
        return QueryTreeOutput(
            data_source_readers={},
            feature_view_readers=prev_step_output.feature_view_readers,
            odfv_input=prev_step_output.odfv_input,
            odfv_output=output_reader,
        )

    def _process_offline_scan_node(self, node: OfflineStoreScanNode, staged_table_name: str) -> None:
        fdw = node.feature_definition_wrapper
        if fdw.has_delta_offline_store:
            reader_params = OfflineStoreReaderParams(delta_table_uri=fdw.materialized_data_path)
            reader = DeltaReader(params=reader_params, fd=fdw, options_providers=self.offline_store_options_providers)
        elif fdw.has_parquet_offline_store:
            reader = ParquetReader(fd=fdw, options_providers=self.offline_store_options_providers)
        else:
            msg = f"Offline store is not configured for FeatureView {fdw.name}"
            raise errors.TectonValidationError(msg)

        compute = SQLCompute.for_dialect(Dialect.DUCKDB)
        self._maybe_register_temp_tables(node.as_ref(), compute)

        if node.entity_filter:
            join_keys_table = compute.run_sql(node.entity_filter.to_sql(), return_dataframe=True).read_all()
            join_keys_filter = {
                join_key: JoinKeyBoundaries(**pc.min_max(column).as_py())
                for join_key, column in zip(join_keys_table.column_names, join_keys_table.columns)
            }
        else:
            join_keys_filter = None

        # We pass join_keys_filter to both OfflineStoreReader,
        # which can leverage this information to select subset of files based on file-level statistics,
        # and to the Compute, which can pushdown this filter to files and read only matching row groups.
        dataset = reader.read(node.partition_time_filter, join_keys_filter)
        s3_options = (
            get_s3_options_for_fd(fdw, self.offline_store_options_providers)
            if fdw.materialized_data_path.startswith("s3://")
            else None
        )

        compute.register_temp_table_from_offline_store(
            staged_table_name, dataset, fdw.materialized_data_path, join_keys_filter, s3_options
        )

    def _process_agg_join(
        self,
        output: QueryTreeOutput,
        compute: QueryTreeCompute,
        qt_node: NodeRef,
        monitor: ComputeMonitor,
    ) -> QueryTreeOutput:
        # TODO(danny): change the "stage" in the StagingNode to be more for the destination stage
        staging_nodes_to_process = get_staging_nodes(qt_node, QueryTreeStep.AGGREGATION)

        if len(staging_nodes_to_process) > 0:
            # There should be a single StagingNode. It is either there for materialization or ODFVs.
            assert len(staging_nodes_to_process) == 1
            readers = self._stage_tables_and_load_pa(
                nodes_to_process=staging_nodes_to_process,
                compute=compute,
                monitor=monitor,
            )
            assert len(readers) == 1
            pa_reader = next(iter(readers.values()))
            return QueryTreeOutput(
                data_source_readers={}, feature_view_readers=output.feature_view_readers, odfv_input=pa_reader
            )

        # There are no StagingNodes, so we can execute the remainder of the query tree.
        output_df_pa = compute.run_sql(qt_node.to_sql(), return_dataframe=True, monitor=monitor)
        return QueryTreeOutput(
            data_source_readers={}, feature_view_readers=output.feature_view_readers, odfv_input=output_df_pa
        )

    def _stage_tables_and_load_pa(
        self,
        nodes_to_process: Dict[str, QueryNode],
        compute: QueryTreeCompute,
        monitor: ComputeMonitor,
    ) -> Dict[str, pyarrow.RecordBatchReader]:
        readers = {}
        for _, node in nodes_to_process.items():
            if isinstance(node, StagingNode):
                name, reader = self._process_staging_node(node, compute, monitor)
                readers[name] = reader
        return readers

    def _process_staging_node(
        self,
        qt_node: StagingNode,
        compute: SQLCompute,
        monitor: ComputeMonitor,
    ) -> Tuple[str, pyarrow.RecordBatchReader]:
        start_time = datetime.now()
        staging_table_name = qt_node.staging_table_name_unique()
        sql_string = qt_node.with_dialect(compute.get_dialect())._to_staging_query_sql()
        reader = compute.run_sql(
            sql_string, return_dataframe=True, expected_output_schema=qt_node.output_schema, monitor=monitor
        )
        staging_done_time = datetime.now()
        if self.is_debug:
            elapsed_staging_time = (staging_done_time - start_time).total_seconds()
            logger.warning(f"STAGE_{staging_table_name}_TIME_SEC: {elapsed_staging_time}")

        return staging_table_name, reader

    def _maybe_register_temp_tables(self, qt_root: NodeRef, compute: SQLCompute) -> None:
        self._dialect_to_temp_table_name = self._dialect_to_temp_table_name or {}

        dialect = compute.get_dialect()
        if dialect not in self._dialect_to_temp_table_name:
            self._dialect_to_temp_table_name[dialect] = set()

        def maybe_register_temp_table(node):
            if isinstance(node, UserSpecifiedDataNode):
                tmp_table_name = node.data._temp_table_name
                if tmp_table_name in self._dialect_to_temp_table_name[dialect]:
                    return
                df = node.data.to_pandas()
                if self.is_debug:
                    logger.warning(
                        f"Registering user specified data {tmp_table_name} to {compute.get_dialect()} with schema:\n{df.dtypes}"
                    )
                compute.register_temp_table_from_pandas(tmp_table_name, df)
                self._dialect_to_temp_table_name[dialect].add(tmp_table_name)

        recurse_query_tree(
            qt_root,
            maybe_register_temp_table,
        )



import functools
import itertools
import logging
import os
import random
import time
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from typing import List
from typing import Optional

import pendulum
from py4j.protocol import Py4JJavaError
from pyspark.sql import Column
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql import functions
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StructType
from pyspark.sql.types import TimestampType

from tecton_core import time_utils as core_time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.offline_store import TIME_PARTITION
from tecton_core.offline_store import PartitionType
from tecton_core.offline_store import _check_supported_offline_store_version
from tecton_core.offline_store import datetime_to_partition_str
from tecton_core.offline_store import partition_col_for_parquet
from tecton_core.offline_store import partition_size_for_delta
from tecton_core.offline_store import partition_size_for_parquet
from tecton_core.offline_store import partition_type_for_delta
from tecton_core.offline_store import timestamp_formats
from tecton_core.offline_store import window_size_seconds
from tecton_core.query_consts import anchor_time
from tecton_proto.data.feature_store__client_pb2 import FeatureStoreFormatVersion
from tecton_spark import time_utils as spark_time_utils


DBRICKS_MULTI_CLUSTER_WRITES_ENABLED = "spark.databricks.delta.multiClusterWrites.enabled"
DBRICKS_RUNTIME_VERSION = "DATABRICKS_RUNTIME_VERSION"

SPARK_GCS_DELTA_LOGSTORE_CLASS = "io.delta.storage.GCSLogStore"

SPARK31_DELTA_LOGSTORE_CLASS = "spark.delta.logStore.class"
SPARK31_DYNAMODB_LOGSTORE_CLASS = "io.delta.storage.DynamoDBLogStore"

SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS = "spark.delta.logStore.s3.impl"
SPARK32_OR_HIGHER_DYNAMODB_LOGSTORE_CLASS = "io.delta.storage.S3DynamoDBLogStore"

logger = logging.getLogger(__name__)


@dataclass
class OfflineStoreWriterParams:
    s3_path: str

    """Whether the anchor column should be stored in the Offline Feature Store regardless of whether it is
    required by the storage layer.

    If this is false the anchor column will be dropped from the stored data if it's not needed by the
    OfflineStoreWriter implementation.
    """
    always_store_anchor_column: bool

    """The column containing the timestamp value used for time-based partitioning"""
    time_column: str

    join_key_columns: List[str]

    is_continuous: bool


def get_offline_store_writer(
    params: OfflineStoreWriterParams, fd: FeatureDefinitionWrapper, version: int, spark: SparkSession
) -> "OfflineStoreWriter":
    """Creates a concrete implementation of OfflineStoreWriter based on fv_config."""
    _check_supported_offline_store_version(fd)
    fv_config = fd.offline_store_config

    case = fv_config.WhichOneof("store_type")
    if case == "delta":
        partition_size = partition_size_for_delta(fd).as_timedelta()
        partition_type = partition_type_for_delta(fd.offline_store_params)
        return DeltaWriter(params, spark, version, partition_size, partition_type)
    elif case == "parquet":
        partition_size = partition_size_for_parquet(fd).as_timedelta()
        partition_col = partition_col_for_parquet(fd)
        return ParquetWriter(params, spark, version, partition_size, partition_col)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    partition_size = partition_size_for_parquet(fd).as_timedelta()
    partition_col = partition_col_for_parquet(fd)
    return ParquetWriter(params, spark, version, partition_size, partition_col)


def get_offline_store_reader(
    spark: SparkSession, fd: FeatureDefinitionWrapper, path: Optional[str] = None
) -> "OfflineStoreReader":
    _check_supported_offline_store_version(fd)
    assert fd.materialization_enabled and fd.writes_to_offline_store
    s3_path = path or fd.materialized_data_path
    version = fd.get_feature_store_format_version

    case = fd.offline_store_config.WhichOneof("store_type")
    if case == "delta":
        partition_size = partition_size_for_delta(fd).as_timedelta()
        partition_type = partition_type_for_delta(fd.offline_store_params)
        return DeltaReader(spark, s3_path, partition_size, partition_type, version)
    elif case == "parquet":
        partition_size = partition_size_for_parquet(fd).as_timedelta()
        partition_col = partition_col_for_parquet(fd)
        return ParquetReader(spark, s3_path, partition_size, partition_col, version)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    partition_size = partition_size_for_parquet(fd).as_timedelta()
    partition_col = partition_col_for_parquet(fd)
    return ParquetReader(spark, s3_path, partition_size, partition_col, version)


class OfflineStoreWriter(ABC):
    """Interface for Offline Feature Store writers."""

    @abstractmethod
    def append_dataframe(self, data_frame: DataFrame) -> None:
        """Append the rows from data_frame to the Store table. Nothing is overwritten."""
        raise NotImplementedError

    @abstractmethod
    def upsert_dataframe(self, data_frame: DataFrame) -> None:
        """Upsert the rows from data_frame to the Store table.

        Rows with matching join keys and time column are overwritten. Other rows are inserted.
        """
        raise NotImplementedError

    @abstractmethod
    def delete_keys(self, data_frame: DataFrame) -> int:
        """Delete rows from the Store table that match the keys inside the data_frame.

        Return number of successfully deleted keys."""
        raise NotImplementedError


class OfflineStoreReader(ABC):
    @abstractmethod
    def read(self, partition_time_limits: pendulum.Period) -> DataFrame:
        """Note that partition_time_limits only applies partition filtering, so you can have records outside it"""
        raise NotImplementedError


class ParquetWriter(OfflineStoreWriter):
    """Parquet implementation of OfflineStoreWriter"""

    def __init__(
        self,
        params: OfflineStoreWriterParams,
        spark: SparkSession,
        version: int,
        partition_size: timedelta,
        partition_col: str,
    ) -> None:
        self._params = params
        self._spark = spark
        self._version = version
        self._partition_size = partition_size
        self._partition_col = partition_col

    def append_dataframe(self, data_frame: DataFrame) -> None:
        if self._partition_col == TIME_PARTITION:
            align_duration = core_time_utils.convert_timedelta_for_version(self._partition_size, self._version)
            aligned_time = _align_timestamp(functions.col(anchor_time()), functions.lit(align_duration))
            data_frame = data_frame.withColumn(TIME_PARTITION, aligned_time)

        data_frame.write.option("partitionOverwriteMode", "dynamic").partitionBy(self._partition_col).parquet(
            self._params.s3_path, mode="overwrite"
        )

    def upsert_dataframe(self, data_frame: DataFrame) -> None:
        raise NotImplementedError()

    def delete_keys(self, data_frame: DataFrame) -> int:
        raise NotImplementedError()


class ParquetReader(OfflineStoreReader):
    def __init__(
        self, spark: SparkSession, path: str, partition_size: timedelta, partition_col: str, version: int
    ) -> None:
        self._spark = spark
        self._path = path
        self._partition_size = partition_size
        self._partition_col = partition_col
        self._version = version

    def read(self, partition_time_limits: Optional[pendulum.Period]) -> DataFrame:
        spark_df = self._spark.read.parquet(self._path)

        # Parquet is partitioned by TIME_PARTITION when is_continuous and ANCHOR_TIME when not.
        # We want to explicitly cast the partition type in case:
        #   `spark.sql.sources.partitionColumnTypeInference.enabled` = "false"

        spark_df = spark_df.withColumn(self._partition_col, functions.col(self._partition_col).cast("long"))

        if partition_time_limits and self._partition_size:
            aligned_start_time = core_time_utils.align_time_downwards(partition_time_limits.start, self._partition_size)
            aligned_end_time = core_time_utils.align_time_downwards(partition_time_limits.end, self._partition_size)
            start_time_epoch = core_time_utils.convert_timestamp_for_version(aligned_start_time, self._version)
            end_time_epoch = core_time_utils.convert_timestamp_for_version(aligned_end_time, self._version)
            partition_col = functions.col(self._partition_col)
            spark_df = spark_df.where((start_time_epoch <= partition_col) & (partition_col <= end_time_epoch))

        return spark_df.drop(TIME_PARTITION)


_EXCEPTION_PACKAGES = {
    "com.databricks.sql.transaction.tahoe",  # Used by Databricks
    "org.apache.spark.sql.delta",  # Used by open source
}

_EXCEPTION_CLASSES = {
    "ConcurrentAppendException",
    "ConcurrentDeleteReadException",
    "ConcurrentDeleteDeleteException",
    "ProtocolChangedException",  # This can occur when two txns create the same table concurrently
}

_RETRYABLE_DELTA_EXCEPTIONS = {
    f"{pkg}.{cls}" for pkg, cls in itertools.product(_EXCEPTION_PACKAGES, _EXCEPTION_CLASSES)
}


def _with_delta_retries(f, max_retries=5):
    """Retries the wrapped function upon Deltalake conflict errors."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        from delta.exceptions import ConcurrentAppendException
        from delta.exceptions import ConcurrentDeleteDeleteException
        from delta.exceptions import ConcurrentDeleteReadException
        from delta.exceptions import ConcurrentTransactionException
        from delta.exceptions import DeltaConcurrentModificationException
        from delta.exceptions import MetadataChangedException
        from delta.exceptions import ProtocolChangedException

        final_exception = None
        for i in range(max_retries):
            try:
                if i > 0:
                    # Add a random delay (with exponential backoff) before the retries to decrease
                    # the chance of recurrent conflicts between the parallel offline store writers.
                    # Possible 10s of seconds of delay is insignificant to the overall job's latency.
                    exponential_coef = 2 ** (i - 1)
                    retry_delay = exponential_coef * random.uniform(0, 1)
                    time.sleep(retry_delay)
                f(*args, **kwargs)
                return
            except Py4JJavaError as e:
                exception_class = e.java_exception.getClass().getCanonicalName()
                if exception_class not in _RETRYABLE_DELTA_EXCEPTIONS:
                    raise e
                final_exception = e
                logger.info(
                    f"Delta transaction failed (attempt {i + 1}/5); retrying",
                    exc_info=True,  # Include information about the exception currently being handled
                )
            except (
                ConcurrentAppendException,
                ConcurrentDeleteDeleteException,
                ConcurrentDeleteReadException,
                ConcurrentTransactionException,
                DeltaConcurrentModificationException,
                MetadataChangedException,
                ProtocolChangedException,
            ) as e:
                final_exception = e
                logger.info(
                    f"Delta transaction failed (attempt {i + 1}/5); retrying",
                    exc_info=True,  # Include information about the exception currently being handled
                )
            except Exception:
                logger.warning("Uncaught exception raised during Delta write", exc_info=True)
                raise
        msg = f"Exceeded maximum Delta transaction retries ({max_retries})"
        raise Exception(msg) from final_exception

    return wrapper


def _assert_safe_delta_write_configuration(spark: SparkSession) -> bool:
    """Asserts that the Spark configuration is such that it is safe to write to Delta concurrently.

    With the Open Source Delta JAR installed (as it is on EMR), writing to a Delta table concurrently with another
    Spark cluster could corrupt the table unless the Delta Logstore class is overridden.

    On Databricks everything is fine as multi-cluster writes are enabled (the default).
    """

    configs = {
        DBRICKS_RUNTIME_VERSION: os.environ.get(DBRICKS_RUNTIME_VERSION, None),
        DBRICKS_MULTI_CLUSTER_WRITES_ENABLED: spark.conf.get(DBRICKS_MULTI_CLUSTER_WRITES_ENABLED, None),
        SPARK31_DELTA_LOGSTORE_CLASS: spark.conf.get(SPARK31_DELTA_LOGSTORE_CLASS, None),
        SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS: spark.conf.get(SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS, None),
    }
    if configs[DBRICKS_RUNTIME_VERSION] and configs[DBRICKS_MULTI_CLUSTER_WRITES_ENABLED] == "true":
        return True

    # either the spark 3.1 or spark 3.2+ DELTA_LOGSTORE_CLASS name can be set
    if configs[SPARK31_DELTA_LOGSTORE_CLASS] in [SPARK31_DYNAMODB_LOGSTORE_CLASS, SPARK_GCS_DELTA_LOGSTORE_CLASS]:
        return True
    if configs[SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS] in [
        SPARK32_OR_HIGHER_DYNAMODB_LOGSTORE_CLASS,
        SPARK_GCS_DELTA_LOGSTORE_CLASS,
    ]:
        return True
    msg = f"Configuration is not safe for concurrent writes: {configs}"
    raise AssertionError(msg)


def _is_dbr_version_greater_than_or_equal_to_14(spark: SparkSession) -> bool:
    # Return False if not on Databricks.
    if DBRICKS_RUNTIME_VERSION not in os.environ:
        return False

    dbr_version = os.environ.get(DBRICKS_RUNTIME_VERSION)

    major_version = dbr_version.split(".")[0]  # "11.3" -> "11"
    try:
        return int(major_version) >= 14
    except ValueError:
        exception_msg = f'DBR Version from "os.environ":{dbr_version} is incorrect.'
        raise ValueError(exception_msg)


class DeltaWriter(OfflineStoreWriter):
    """DeltaLake implementation of OfflineStoreWriter"""

    def __init__(
        self,
        params: OfflineStoreWriterParams,
        spark: SparkSession,
        version: int,
        partition_size: timedelta,
        partition_type: PartitionType,
    ) -> None:
        self._params = params
        self._spark = spark
        self._version = version
        self._partition_size = partition_size
        self._partition_type = partition_type
        self._metadata_writer = DeltaMetadataWriter(spark)
        if not spark.conf.get("spark.databricks.delta.commitInfo.userMetadata"):
            msg = "Expected spark.databricks.delta.commitInfo.userMetadata to be set for delta writes"
            raise AssertionError(msg)

    def append_dataframe(self, data_frame: DataFrame) -> None:
        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)
        self._append_dataframe(data_frame)

    def upsert_dataframe(self, data_frame):
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)

        table = DeltaTable.forPath(self._spark, self._params.s3_path)

        base = table.toDF().alias("base")
        updates = data_frame.alias("updates")

        # Build a condition which matches on all join keys, the timestamp, and the time partition column. The time
        # partition column is not needed for correctness, but it allows some files to be skipped by Delta.
        all_match_keys = [self._params.time_column, TIME_PARTITION, *self._params.join_key_columns]
        key_matches = [base[k] == updates[k] for k in all_match_keys]
        match_condition = functools.reduce(lambda l, r: l & r, key_matches)

        @_with_delta_retries
        def _execute():
            table.merge(updates, match_condition).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

        _execute()

    def delete_keys(self, data_frame: DataFrame) -> int:
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        deltaTable = DeltaTable.forPath(self._spark, self._params.s3_path)
        query = ""
        columns = data_frame.columns
        for column in columns:
            if query:
                query = query + " AND "
            query = query + "t." + column + " = k." + column

        @_with_delta_retries
        def _execute():
            deltaTable.alias("t").merge(data_frame.alias("k"), query).whenMatchedDelete().execute()

        @_with_delta_retries
        def _vacuum():
            deltaTable.vacuum()

        _execute()
        _vacuum()

        last_operation = deltaTable.history(1).collect()
        if not last_operation:
            return 0

        return int(last_operation[0].operationMetrics.get("numTargetRowsDeleted", 0))

    def overwrite_dataframe_in_tile(self, data_frame: DataFrame, time_range: pendulum.Period) -> None:
        df_to_write = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, df_to_write.schema)

        replace_predicate = self._in_range(time_range)

        @_with_delta_retries
        def _execute():
            df_to_write.write.format("delta").mode("overwrite").partitionBy(TIME_PARTITION).option(
                "replaceWhere", replace_predicate
            ).save(self._params.s3_path)

        _assert_safe_delta_write_configuration(self._spark)
        _execute()

    def _in_range(self, time_range: pendulum.Period) -> str:
        if self._params.time_column == anchor_time():
            assert self._params.always_store_anchor_column
            start_time = core_time_utils.convert_timestamp_for_version(time_range.start, self._version)
            end_time = core_time_utils.convert_timestamp_for_version(time_range.end, self._version)
            time_col_predicate = (
                f"{self._params.time_column} >= {start_time} AND {self._params.time_column} < {end_time}"
            )
        else:
            start_time_str = time_range.start.strftime("%Y-%m-%d %H:%M:%S")
            end_time_str = time_range.end.strftime("%Y-%m-%d %H:%M:%S")
            time_col_predicate = f"{self._params.time_column} >= timestamp('{start_time_str}') AND {self._params.time_column} < timestamp('{end_time_str}')"

        partition_aligned_start = core_time_utils.align_time_downwards(time_range.start, self._partition_size)
        partition_aligned_end = core_time_utils.align_time_downwards(time_range.end, self._partition_size)
        if self._partition_type == PartitionType.EPOCH:
            start_partition_epoch = core_time_utils.convert_timestamp_for_version(
                partition_aligned_start, self._version
            )
            end_partition_epoch = core_time_utils.convert_timestamp_for_version(partition_aligned_end, self._version)
            partition_predicate = f"{TIME_PARTITION} BETWEEN {start_partition_epoch} AND {end_partition_epoch}"
        else:
            # TODO (vitaly): test that this works for hour partitions (rare case)
            start_partition_str = datetime_to_partition_str(partition_aligned_start, self._partition_size)
            end_partition_str = datetime_to_partition_str(partition_aligned_end, self._partition_size)
            partition_predicate = (
                f"{TIME_PARTITION} BETWEEN timestamp('{start_partition_str}') AND timestamp('{end_partition_str}')"
            )

        return f"{time_col_predicate} AND {partition_predicate}"

    def _add_partition(self, data_frame: DataFrame) -> DataFrame:
        """Adds the time_partition column and drops the _anchor_time column if needed."""
        partition = self._timestamp_to_partition_column(data_frame)
        data_frame = data_frame.withColumn(TIME_PARTITION, partition)
        if not self._params.always_store_anchor_column:
            data_frame = data_frame.drop(anchor_time())
        return data_frame

    def _ensure_table_exists(self, spark: SparkSession, schema: StructType) -> None:
        """Ensures that the table exists with the given schema.

        Some operations (including merge) fail when the table doesn't already exist. Others (append) can have conflicts
        where they wouldn't normally when they also create a new table. This function ensures neither will happen.
        """
        df = spark.createDataFrame([], schema)  # DF with 0 rows
        self._append_dataframe(df)

        # Manifest files are not supported on GCS
        if not self._params.s3_path.startswith("gs://"):
            # we set auto manifest so each job generates its own manifest (necessary for athena retrieval)
            self._metadata_writer.set_table_property(
                self._params.s3_path, "delta.compatibility.symlinkFormatManifest.enabled", "true"
            )

    @_with_delta_retries
    def _append_dataframe(self, df: DataFrame) -> None:
        _assert_safe_delta_write_configuration(self._spark)
        df_writer = df.write.partitionBy(TIME_PARTITION).format("delta").mode("append")
        # For DBR 14+, Deletion Vectors are auto-enabled.
        # Refer to this section on why we must disable deletion vectors:
        # https://www.notion.so/tecton/RFC-Support-DLT-Table-as-Data-Source-7ddf14a8ace04b03ba91b2e3f7db03bf?pvs=4#f520cb9f406d4c6f9fc743dc4f399607
        if _is_dbr_version_greater_than_or_equal_to_14(self._spark):
            df_writer.option("delta.enableDeletionVectors", "false")
        df_writer.save(self._params.s3_path)

    def _timestamp_to_partition_column(self, df: DataFrame) -> Column:
        # For some insane reason from_unixtime returns a timestamp in the session timezone, so it's pretty annoying to
        # convert a unix time to a formatted UTC timestamp unless the session is set to UTC. This only runs in
        # materialization so we can just assert that that's the case.
        tz = df.sql_ctx.sparkSession.conf.get("spark.sql.session.timeZone")
        if tz not in {"UTC", "Etc/UTC", "GMT"}:
            msg = f"spark.sql.session.timeZone must be UTC, not {tz}"
            raise AssertionError(msg)

        time_col = self._params.time_column
        time_column_type = df.schema[time_col].dataType
        allowed_types = {IntegerType(), TimestampType(), LongType()}
        if time_column_type not in allowed_types:
            msg = f"timestamp column must be one of {allowed_types}, not {time_column_type}"
            raise AssertionError(msg)

        if self._partition_type == PartitionType.EPOCH:
            assert (
                self._version >= FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS
            ), f"FeatureStoreFormateVersion {self._version} is not supported. PartitionType.EPOCH on delta must be in nanoseconds."
            partition_size_nanoseconds = core_time_utils.convert_timedelta_for_version(
                self._partition_size, self._version
            )
            time_val_nanoseconds = spark_time_utils.convert_timestamp_to_epoch(functions.col(time_col), self._version)
            aligned_time_partition_val_nanoseconds = _align_timestamp(time_val_nanoseconds, partition_size_nanoseconds)
            return aligned_time_partition_val_nanoseconds

        time_val = functions.col(time_col).cast(LongType())
        if time_col == anchor_time():
            time_val = spark_time_utils.convert_epoch_to_datetime(functions.col(time_col), self._version).cast(
                LongType()
            )
        aligned_time_partition_val_seconds = _align_timestamp(time_val, window_size_seconds(self._partition_size))
        aligned_time_partition_val_string = functions.from_unixtime(aligned_time_partition_val_seconds)
        partition_format_for_date_str = timestamp_formats(self._partition_size).spark_format
        return functions.date_format(
            aligned_time_partition_val_string.cast(TimestampType()), partition_format_for_date_str
        )


class DeltaMetadataWriter:
    def __init__(self, spark: SparkSession) -> None:
        self._spark = spark

    @_with_delta_retries
    def generate_symlink_manifest(self, path: str) -> None:
        _assert_safe_delta_write_configuration(self._spark)
        # we need spark_catalog in cases where the data source switches catalogs
        self._spark.sql(f"GENERATE symlink_format_manifest FOR TABLE spark_catalog.delta.`{path}`")

    @_with_delta_retries
    def set_table_property(self, path, key, val):
        _assert_safe_delta_write_configuration(self._spark)
        # we need spark_catalog in cases where the data source switches catalogs
        existing_tbl_properties = self._spark.sql(f"show tblproperties spark_catalog.delta.`{path}`").collect()
        # tblproperties are case sensitive
        has_tbl_property = any(key == r.key and val == r.value for r in existing_tbl_properties)
        # we only set it if not already set to avoid delta conflicts
        if not has_tbl_property:
            self._spark.sql(f"ALTER TABLE spark_catalog.delta.`{path}` SET TBLPROPERTIES({key}={val})")

    @_with_delta_retries
    def optimize_execute_compaction(self, path):
        _assert_safe_delta_write_configuration(self._spark)
        self._spark.sql(f"OPTIMIZE '{path}'")

    @_with_delta_retries
    def vacuum(self, path):
        _assert_safe_delta_write_configuration(self._spark)
        self._spark.sql(f"VACUUM '{path}'")


class DeltaReader(OfflineStoreReader):
    def __init__(
        self,
        spark: SparkSession,
        path: str,
        partition_size: timedelta,
        partition_type: PartitionType,
        feature_store_format_version: int,
    ) -> None:
        self._spark = spark
        self._path = path
        self._partition_size = partition_size
        self._partition_type = partition_type
        self._feature_store_format_version = feature_store_format_version

    def read(self, partition_time_limits: Optional[pendulum.Period]) -> DataFrame:
        spark_df = self._spark.read.format("delta").load(self._path)

        if partition_time_limits is None or not self._partition_size:
            return spark_df.drop(TIME_PARTITION)

        # Whenever the partition filtering logic is changed, also make sure the changes are applied to the sql based
        # version in query/nodes.py

        # Delta is always partitioned by TIME_PARTITION. We want to explicitly cast to it's correct type in case:
        #   `spark.sql.sources.partitionColumnTypeInference.enabled` = "false"
        aligned_start_time = core_time_utils.align_time_downwards(partition_time_limits.start, self._partition_size)
        aligned_end_time = core_time_utils.align_time_downwards(partition_time_limits.end, self._partition_size)

        if self._partition_type == PartitionType.DATE_STR:
            partition_col = functions.col(TIME_PARTITION).cast("timestamp")
            start_partition = datetime_to_partition_str(aligned_start_time, self._partition_size)
            end_partition = datetime_to_partition_str(aligned_end_time, self._partition_size)
        elif self._partition_type == PartitionType.EPOCH:
            partition_col = functions.col(TIME_PARTITION).cast("long")
            start_partition = core_time_utils.convert_timestamp_for_version(
                aligned_start_time, self._feature_store_format_version
            )
            end_partition = core_time_utils.convert_timestamp_for_version(
                aligned_end_time, self._feature_store_format_version
            )
        else:
            msg = f"Invalid partition type for Delta: {self._partition_type}"
            raise AssertionError(msg)

        spark_df = spark_df.where((start_partition <= partition_col) & (partition_col <= end_partition))

        return spark_df.drop(TIME_PARTITION)


def _align_timestamp(int_timestamp_col, window_size):
    return int_timestamp_col - (int_timestamp_col % window_size)



from __future__ import absolute_import


from pyspark.cloudpickle.cloudpickle import *  # noqa
from pyspark.cloudpickle.cloudpickle_fast import CloudPickler, dumps, dump  # noqa

__version__ = '1.5.0'



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/dataobs/validation_task__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import fco_locator__client_pb2 as tecton__proto_dot_common_dot_fco__locator____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.dataobs import expectation__client_pb2 as tecton__proto_dot_dataobs_dot_expectation____client__pb2
from tecton_proto.dataobs import validation__client_pb2 as tecton__proto_dot_dataobs_dot_validation____client__pb2
from tecton_proto.dataobs import validation_task_params__client_pb2 as tecton__proto_dot_dataobs_dot_validation__task__params____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/dataobs/validation_task__client.proto\x12\x14tecton_proto.dataobs\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a-tecton_proto/common/fco_locator__client.proto\x1a$tecton_proto/common/id__client.proto\x1a.tecton_proto/dataobs/expectation__client.proto\x1a-tecton_proto/dataobs/validation__client.proto\x1a\x39tecton_proto/dataobs/validation_task_params__client.proto\"\xc3\x05\n\x0eValidationTask\x12\x43\n\x11validation_job_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fvalidationJobId\x12S\n\x14\x66\x65\x61ture_view_locator\x18\x02 \x01(\x0b\x32!.tecton_proto.common.IdFcoLocatorR\x12\x66\x65\x61tureViewLocator\x12X\n\x13metric_expectations\x18\x03 \x03(\x0b\x32\'.tecton_proto.dataobs.MetricExpectationR\x12metricExpectations\x12H\n\x12\x66\x65\x61ture_start_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12\x45\n\ttask_type\x18\x08 \x01(\x0e\x32(.tecton_proto.dataobs.ValidationTaskTypeR\x08taskType\x12\x33\n\x07timeout\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationR\x07timeout\x12V\n\x12\x64ynamo_data_source\x18\x06 \x01(\x0b\x32&.tecton_proto.dataobs.DynamoDataSourceH\x00R\x10\x64ynamoDataSource\x12J\n\x0es3_data_source\x18\x07 \x01(\x0b\x32\".tecton_proto.dataobs.S3DataSourceH\x00R\x0cs3DataSourceB\r\n\x0b\x64\x61ta_source\"\x9f\x04\n\x14ValidationTaskResult\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12\x45\n\x12\x66\x65\x61ture_package_id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61turePackageId\x12\x43\n\x11validation_job_id\x18\x03 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fvalidationJobId\x12H\n\x12\x66\x65\x61ture_start_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12\x41\n\x07results\x18\x06 \x03(\x0b\x32\'.tecton_proto.dataobs.ExpectationResultR\x07results\x12\x45\n\x07metrics\x18\x07 \x01(\x0b\x32+.tecton_proto.dataobs.ValidationTaskMetricsR\x07metrics\x12\x43\n\x0fvalidation_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0evalidationTime\"\xbc\x01\n\x15ValidationTaskMetrics\x12(\n\x10metric_rows_read\x18\x01 \x01(\rR\x0emetricRowsRead\x12*\n\x11\x66\x65\x61ture_rows_read\x18\x02 \x01(\rR\x0f\x66\x65\x61tureRowsRead\x12M\n\x15query_execution_times\x18\x03 \x03(\x0b\x32\x19.google.protobuf.DurationR\x13queryExecutionTimes*^\n\x12ValidationTaskType\x12 \n\x1cVALIDATION_TASK_TYPE_UNKNOWN\x10\x00\x12&\n\"VALIDATION_TASK_TYPE_BATCH_METRICS\x10\x01\x42\x16\n\x12\x63om.tecton.dataobsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.dataobs.validation_task__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.dataobsP\001'
  _VALIDATIONTASKTYPE._serialized_start=1827
  _VALIDATIONTASKTYPE._serialized_end=1921
  _VALIDATIONTASK._serialized_start=381
  _VALIDATIONTASK._serialized_end=1088
  _VALIDATIONTASKRESULT._serialized_start=1091
  _VALIDATIONTASKRESULT._serialized_end=1634
  _VALIDATIONTASKMETRICS._serialized_start=1637
  _VALIDATIONTASKMETRICS._serialized_end=1825
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This file is ported from spark/util/StatCounter.scala

import copy
import math

try:
    from numpy import maximum, minimum, sqrt
except ImportError:
    maximum = max
    minimum = min
    sqrt = math.sqrt


class StatCounter(object):

    def __init__(self, values=None):
        if values is None:
            values = list()
        self.n = 0    # Running count of our values
        self.mu = 0.0  # Running mean of our values
        self.m2 = 0.0  # Running variance numerator (sum of (x - mean)^2)
        self.maxValue = float("-inf")
        self.minValue = float("inf")

        for v in values:
            self.merge(v)

    # Add a value into this StatCounter, updating the internal statistics.
    def merge(self, value):
        delta = value - self.mu
        self.n += 1
        self.mu += delta / self.n
        self.m2 += delta * (value - self.mu)
        self.maxValue = maximum(self.maxValue, value)
        self.minValue = minimum(self.minValue, value)

        return self

    # Merge another StatCounter into this one, adding up the internal statistics.
    def mergeStats(self, other):
        if not isinstance(other, StatCounter):
            raise Exception("Can only merge Statcounters!")

        if other is self:  # reference equality holds
            self.merge(copy.deepcopy(other))  # Avoid overwriting fields in a weird order
        else:
            if self.n == 0:
                self.mu = other.mu
                self.m2 = other.m2
                self.n = other.n
                self.maxValue = other.maxValue
                self.minValue = other.minValue

            elif other.n != 0:
                delta = other.mu - self.mu
                if other.n * 10 < self.n:
                    self.mu = self.mu + (delta * other.n) / (self.n + other.n)
                elif self.n * 10 < other.n:
                    self.mu = other.mu - (delta * self.n) / (self.n + other.n)
                else:
                    self.mu = (self.mu * self.n + other.mu * other.n) / (self.n + other.n)

                self.maxValue = maximum(self.maxValue, other.maxValue)
                self.minValue = minimum(self.minValue, other.minValue)

                self.m2 += other.m2 + (delta * delta * self.n * other.n) / (self.n + other.n)
                self.n += other.n
        return self

    # Clone this StatCounter
    def copy(self):
        return copy.deepcopy(self)

    def count(self):
        return int(self.n)

    def mean(self):
        return self.mu

    def sum(self):
        return self.n * self.mu

    def min(self):
        return self.minValue

    def max(self):
        return self.maxValue

    # Return the variance of the values.
    def variance(self):
        if self.n == 0:
            return float('nan')
        else:
            return self.m2 / self.n

    #
    # Return the sample variance, which corrects for bias in estimating the variance by dividing
    # by N-1 instead of N.
    #
    def sampleVariance(self):
        if self.n <= 1:
            return float('nan')
        else:
            return self.m2 / (self.n - 1)

    # Return the standard deviation of the values.
    def stdev(self):
        return sqrt(self.variance())

    #
    # Return the sample standard deviation of the values, which corrects for bias in estimating the
    # variance by dividing by N-1 instead of N.
    #
    def sampleStdev(self):
        return sqrt(self.sampleVariance())

    def asDict(self, sample=False):
        """Returns the :class:`StatCounter` members as a ``dict``.

        Examples
        --------
        >>> sc.parallelize([1., 2., 3., 4.]).stats().asDict()
        {'count': 4L,
         'max': 4.0,
         'mean': 2.5,
         'min': 1.0,
         'stdev': 1.2909944487358056,
         'sum': 10.0,
         'variance': 1.6666666666666667}
        """
        return {
            'count': self.count(),
            'mean': self.mean(),
            'sum': self.sum(),
            'min': self.min(),
            'max': self.max(),
            'stdev': self.stdev() if sample else self.sampleStdev(),
            'variance': self.variance() if sample else self.sampleVariance()
        }

    def __repr__(self):
        return ("(count: %s, mean: %s, stdev: %s, max: %s, min: %s)" %
                (self.count(), self.mean(), self.stdev(), self.max(), self.min()))



import inspect
import logging
import re
import site
import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set

from tecton_core import conf
from tecton_proto.args import repo_metadata__client_pb2 as repo_metadata_pb2
from tecton_proto.common import id__client_pb2 as id_pb2


# TODO(brian): refactor functions/callers here to use the RepoData class and
# remove these duplicative helpers.

logger = logging.getLogger(__name__)


@dataclass
class RepoData:
    # paths and file set are different representations of the same data
    root: str
    paths: List[Path] = field(default_factory=list)
    file_set: Set[str] = field(default_factory=set)


_repo_data: Optional[RepoData] = None


def get_repo_data() -> Optional[RepoData]:
    return _repo_data


def set_repo_data(repo_data: RepoData) -> None:
    global _repo_data
    _repo_data = repo_data


# Prepare a repo for tecton objects to be processed.
# If file_in_repo is not set, expects the current directory to be inside the feature repo.
def ensure_prepare_repo(file_in_repo: Optional[str] = None) -> None:
    repo_data = get_repo_data()
    if repo_data:
        # repo is already prepared
        return
    root = _maybe_get_repo_root(file_in_repo)
    if root is None:
        msg = "Feature repository root not found. Run `tecton init` to set it."
        raise Exception(msg)
    paths = get_repo_files(root)
    file_set = {str(f) for f in paths}
    set_repo_data(RepoData(paths=paths, file_set=file_set, root=root))


def repo_files() -> List[Path]:
    repo_data = get_repo_data()
    if repo_data is None:
        msg = "Repo is not prepared"
        raise Exception(msg)
    return repo_data.paths


def repo_files_set() -> Set[str]:
    repo_data = get_repo_data()
    if repo_data is None:
        msg = "Repo is not prepared"
        raise Exception(msg)
    return repo_data.file_set


def repo_root() -> str:
    repo_data = get_repo_data()
    if repo_data is None:
        msg = "Repo is not prepared"
        raise Exception(msg)
    return repo_data.root


def _fake_init_for_testing(root: str = "") -> None:
    fake_root = Path(root)
    set_repo_data(RepoData(paths=[fake_root], file_set={str(fake_root)}, root=str(fake_root)))


# Finds the repo root of a given python file (a parent directory containing ".tecton")
# If a file is not passed in, searches the current directory's parents.
# Returns None if the root is not found
def _maybe_get_repo_root(file_in_repo: Optional[str] = None) -> Optional[str]:
    repo_data = get_repo_data()
    if repo_data is not None and repo_data.root:
        return repo_data.root

    if file_in_repo is None:
        d = Path().resolve()
    else:
        d = Path(file_in_repo)

    while d.parent != d and d != Path.home():
        tecton_cfg = d / Path(".tecton")
        if tecton_cfg.exists():
            # NOTE: we are not caching this since traversing up the file system
            # should be quick and calls to this should be infrequent.
            #
            # Previous implementations did cache this in our `_repo_data` and
            # that caused a bug for `tecton test`.
            return str(d)
        d = d.parent
    return None


def is_file_in_a_tecton_repo(file: str) -> bool:
    """Determines if the file is in a Tecton repo, i.e. a parent directory contains a .tecton file."""
    d = Path(file)
    while d.parent != d and d != Path.home():
        tecton_cfg = d / Path(".tecton")
        if tecton_cfg.exists():
            return True
        d = d.parent
    return False


def _get_ignored_files(repo_root: Path) -> Set[Path]:
    import pathspec

    ignorefile = repo_root / Path(".tectonignore")
    if not ignorefile.exists():
        return set()

    with open(ignorefile, "r") as f:
        spec = pathspec.PathSpec.from_lines("gitwildmatch", f)
        return {repo_root / file_name for file_name in spec.match_tree(str(repo_root))}


def get_repo_files(root: str, suffixes: Iterable[str] = (".py", ".yml", "yaml")) -> List[Path]:
    root_path = Path(root)
    repo_files = [p.resolve() for p in root_path.glob("**/*") if p.suffix in suffixes]

    # Ignore virtualenv directory if any, typically you'd have /some/path/bin/python as an
    # interpreter, so we want to skip anything under /some/path/
    if sys.executable:
        python_dir = Path(sys.executable).parent.parent

        # we might be dealing with virtualenv
        if root_path.resolve() in python_dir.parents:
            repo_files = [p for p in repo_files if python_dir not in p.parents]

    # NOTE: we added this behind a flag so that customers can disable this behavior if their environment
    # places their repo in a hidden directory as part of its execution. This happened with a customer
    # that used Bazel which configured the repo under the `~/.cache/bazel/...` directory.
    # https://tecton.atlassian.net/browse/CS-3509
    if conf.get_bool("TECTON_REPO_IGNORE_ALL_HIDDEN_DIRS"):
        # Filter out files under hidden dirs starting with "/." for PosixPath or "\." for WindowsPath
        repo_files = list(filter(lambda p: "/." not in str(p) and "\." not in str(p), repo_files))

    # Filter out files that match glob expressions in .tectonignore
    ignored_files = _get_ignored_files(root_path)
    filtered_files = [p for p in repo_files if p not in ignored_files]

    return filtered_files


def construct_fco_source_info(fco_id: id_pb2.Id) -> repo_metadata_pb2.SourceInfo:
    """Get the SourceInfo (file name and line number) for an FCO.

    How it works:
    - This function assumed it is being called from the constructor of an FCO
    - inspect.stack() returns the call stack (starting with this function)
    - Walk up the stack frames until the first file within a tecton repo (a child of .tecton) is found
    - The first valid tecton repo file is considered the filename of the FCO.
    """
    from tecton_core.repo_file_handler import _maybe_get_repo_root

    # Matches frame strings such as "<string>"
    SKIP_FRAME_REGEX = re.compile("\<.*\>")

    source_info = repo_metadata_pb2.SourceInfo(fco_id=fco_id)
    repo_root = _maybe_get_repo_root()
    if not repo_root:
        return source_info

    # 'getsitepackages' and 'getusersitepackages' are not avaiable in some python envs such as EMR notebook with
    # Python 3.7.
    if not (hasattr(site, "getsitepackages") and hasattr(site, "getusersitepackages")):
        logger.warn(
            "Python 'site' pakcage doesn't contain 'getsitepackages' or 'getusersitepackages' methods. SourceInfo is not going to be populated."
        )
        return source_info

    excluded_site_pkgs = [*site.getsitepackages(), site.getusersitepackages()]

    frames = inspect.stack()
    repo_root_path = Path(repo_root)
    for frame in frames:
        if SKIP_FRAME_REGEX.match(frame.frame.f_code.co_filename) is not None:
            continue
        frame_path = Path(frame.frame.f_code.co_filename).resolve()
        if not frame_path.exists():
            continue
        if any(pkg in frame.frame.f_code.co_filename for pkg in excluded_site_pkgs):
            # This filtering is needed in case `tecton` is installed using a virtual
            # environment that's created *inside* the repo root. Without this check,
            # Tecton SDK files would incorrectly be considered a valid repo files
            # and would be listed as the FCO's source filename.
            continue
        if repo_root_path in frame_path.parents:
            rel_filename = frame_path.relative_to(repo_root_path)
            source_info.source_lineno = str(frame.lineno)
            source_info.source_filename = str(rel_filename)
            break
    return source_info



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1720474544', 'BUILD_TIMESTAMP': '1720474544', 'FORMATTED_DATE': '2024 Jul 08 21 35 44 Mon', 'GIT_COMMIT': '23555ae47e66c70b59532fd618d87d6f83a4a925', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-medium-5bcd4bb9c9-9rw7t', 'BUILD_USER': 'buildkite-agent'}



from datetime import datetime
from typing import TYPE_CHECKING
from typing import Optional

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession


if TYPE_CHECKING:
    from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


_active_metrics_collector = None


def get_active_metrics_collector():
    if _active_metrics_collector is None:
        return NoopMetricsCollector()

    return _active_metrics_collector


class MetricsCollector:
    def observe(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError

    def publish(self):
        raise NotImplementedError


class NoopMetricsCollector(MetricsCollector):
    def observe(self, df: DataFrame) -> DataFrame:
        return df

    def publish(self):
        pass


class SparkMetricsCollector(MetricsCollector):
    def __init__(self, jvm_collector):
        self._jvm_collector = jvm_collector

    def observe(self, df: DataFrame) -> DataFrame:
        new_jdf = self._jvm_collector.observe(df._jdf)
        return DataFrame(new_jdf, df.sql_ctx)

    def publish(self):
        self._jvm_collector.publish()


def create_feature_metrics_collector(
    spark: SparkSession,
    params: "MaterializationTaskParams",
    feature_start_time: Optional[datetime] = None,
    feature_end_time: Optional[datetime] = None,
) -> MetricsCollector:
    global _active_metrics_collector

    if not params.HasField("data_observability_config"):
        return NoopMetricsCollector()

    config = params.data_observability_config
    if not config.enabled:
        return NoopMetricsCollector()

    # TODO(dataobs): it seems we always used the default datetime for FT ingestion, which is 0
    # if it's intentional, it should be more explicit
    feature_start_time = feature_start_time or datetime.fromtimestamp(0)
    feature_end_time = feature_end_time or datetime.fromtimestamp(0)

    jvm_collector = spark._jvm.com.tecton.dataobs.spark.MetricsCollector.fromMaterializationTaskParams(
        params.SerializeToString(),
        int(feature_start_time.timestamp()),
        int(feature_end_time.timestamp()),
        spark._jsparkSession,
    )
    _active_metrics_collector = SparkMetricsCollector(jvm_collector)
    return _active_metrics_collector



import base64
import logging
from collections import defaultdict
from types import ModuleType
from typing import List

from google.protobuf.message import Message

from tecton_core import http
from tecton_core.metadata_service_impl import base_stub
from tecton_core.metadata_service_impl import error_lib
from tecton_core.metadata_service_impl import providers
from tecton_core.metadata_service_impl.response import MDSResponse
from tecton_core.metadata_service_impl.service_calls import GrpcCall
from tecton_core.metadata_service_impl.service_calls import get_method_name_to_grpc_call


logger = logging.getLogger(__name__)

requests_session = http.session()


class PureHTTPStub(base_stub.BaseStub):
    def __init__(self, request_provider: providers.RequestProvider, grpc_service_modules: List[ModuleType]) -> None:
        """
        PureHTTPStub mimics the interface exposed by a gRPC client for MetadataService.
        Callers can invoke a method like stub.GetFeatureView and it's transformed for an HTTP
        request using _InternalHTTPStub in the __getattr__ method below.
        """
        self._stub_obj = _InternalHTTPStub(request_provider)
        self._method_name_to_grpc_call = get_method_name_to_grpc_call(grpc_service_modules)

    def __getattr__(self, method_name):
        """
        Transforms methods called directly on PureHTTPStub to _InternalHTTPStub requests.
        E.g. PureHTTPStub::SomeMethod(request) is transformed to _InternalHTTPStub::execute('SomeMethod', request).

        An AttributeError is raised if the method invoked does not match a MetadataService RPC method.
        """
        if method_name not in self._method_name_to_grpc_call:
            msg = f"Nonexistent MetadataService method: {method_name}"
            raise AttributeError(msg)

        def method(request: Message, timeout_sec: float = 300.0) -> MDSResponse:
            return self._stub_obj.execute(self._method_name_to_grpc_call[method_name], request, timeout_sec)

        return method

    def close(self):
        pass


class _InternalHTTPStub:
    def __init__(self, request_provider: providers.RequestProvider) -> None:
        self.request_provider = request_provider

    def execute(self, grpc_call: GrpcCall, request: Message, timeout_sec: float) -> MDSResponse:
        """
        :param grpc_call: gRPC call object representation
        :param request: Request proto.
        :param timeout_sec: timeout for request in seconds

        :return: Response proto.
        """
        json_request = {}
        json_request["method"] = grpc_call.method
        json_request["metadata"] = self.request_provider.request_headers()
        json_request["request"] = base64.encodebytes(grpc_call.request_serializer(request)).decode("utf-8")

        request_url = self.request_provider.request_url()
        response = requests_session.post(request_url, json=json_request, timeout=timeout_sec)
        response.raise_for_status()
        body = response.json()

        code = body["status"]["code"]
        if code != error_lib.gRPCStatus.OK.value:
            details = body["status"]["detail"]
            error_lib.raise_for_grpc_status(code, details, request_url, self.request_provider)

        response_bytes = base64.decodebytes(body["response"].encode("utf-8"))
        response_proto = grpc_call.response_deserializer(response_bytes)
        metadata = body.get("metadata", "")
        return MDSResponse(response_proto, defaultdict(str, metadata))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/online_store_writer/copier__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.data import feature_view__client_pb2 as tecton__proto_dot_data_dot_feature__view____client__pb2
from tecton_proto.online_store_writer import config__client_pb2 as tecton__proto_dot_online__store__writer_dot_config____client__pb2
from tecton_proto.snowflake import snowflake_credentials__client_pb2 as tecton__proto_dot_snowflake_dot_snowflake__credentials____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/online_store_writer/copier__client.proto\x12 tecton_proto.online_store_writer\x1a\x1fgoogle/protobuf/timestamp.proto\x1a$tecton_proto/common/id__client.proto\x1a,tecton_proto/data/feature_view__client.proto\x1a\x35tecton_proto/online_store_writer/config__client.proto\x1a:tecton_proto/snowflake/snowflake_credentials__client.proto\"\xcd\x07\n\x18OnlineStoreCopierRequest\x12\x8b\x01\n!online_store_writer_configuration\x18\x01 \x01(\x0b\x32@.tecton_proto.online_store_writer.OnlineStoreWriterConfigurationR\x1eonlineStoreWriterConfiguration\x12\x41\n\x0c\x66\x65\x61ture_view\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FeatureViewR\x0b\x66\x65\x61tureView\x12^\n\x1fmaterialization_task_attempt_id\x18\x05 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x1cmaterializationTaskAttemptId\x12\x19\n\x08task_key\x18\t \x01(\tR\x07taskKey\x12#\n\rattempt_index\x18\x08 \x01(\x05R\x0c\x61ttemptIndex\x12\x17\n\x07sqs_url\x18\n \x01(\tR\x06sqsUrl\x12\x91\x01\n#online_store_metadata_configuration\x18\x0b \x01(\x0b\x32\x42.tecton_proto.online_store_writer.OnlineStoreMetadataConfigurationR onlineStoreMetadataConfiguration\x12\x43\n\x1e\x65nable_new_consumption_metrics\x18\x0c \x01(\x08R\x1b\x65nableNewConsumptionMetrics\x12\x65\n\x13object_copy_request\x18\x06 \x01(\x0b\x32\x33.tecton_proto.online_store_writer.ObjectCopyRequestH\x00R\x11objectCopyRequest\x12k\n\x15status_update_request\x18\x07 \x01(\x0b\x32\x35.tecton_proto.online_store_writer.StatusUpdateRequestH\x00R\x13statusUpdateRequest\x12^\n\x10\x64\x65letion_request\x18\r \x01(\x0b\x32\x31.tecton_proto.online_store_writer.DeletionRequestH\x00R\x0f\x64\x65letionRequestB\x0e\n\x0crequest_typeJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\"3\n\x07S3Stage\x12\x16\n\x06\x62ucket\x18\x01 \x01(\tR\x06\x62ucket\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\"6\n\x08GCSStage\x12\x16\n\x06\x62ucket\x18\x01 \x01(\tR\x06\x62ucket\x12\x12\n\x04\x62lob\x18\x02 \x01(\tR\x04\x62lob\"\xc6\x01\n\x16SnowflakeInternalStage\x12@\n\x1csnowflake_account_identifier\x18\x01 \x01(\tR\x1asnowflakeAccountIdentifier\x12N\n\x0b\x63redentials\x18\x02 \x01(\x0b\x32,.tecton_proto.snowflake.SnowflakeCredentialsR\x0b\x63redentials\x12\x1a\n\x08location\x18\x03 \x01(\tR\x08location\",\n\x0eLocalFileStage\x12\x1a\n\x08location\x18\x01 \x01(\tR\x08location\"\x8b\x04\n\x11ObjectCopyRequest\x12\x46\n\x08s3_stage\x18\x05 \x01(\x0b\x32).tecton_proto.online_store_writer.S3StageH\x00R\x07s3Stage\x12t\n\x18snowflake_internal_stage\x18\x06 \x01(\x0b\x32\x38.tecton_proto.online_store_writer.SnowflakeInternalStageH\x00R\x16snowflakeInternalStage\x12\\\n\x10local_file_stage\x18\x07 \x01(\x0b\x32\x30.tecton_proto.online_store_writer.LocalFileStageH\x00R\x0elocalFileStage\x12I\n\tgcs_stage\x18\t \x01(\x0b\x32*.tecton_proto.online_store_writer.GCSStageH\x00R\x08gcsStage\x12\x1b\n\tskip_rows\x18\x03 \x01(\x03R\x08skipRows\x12X\n\x0ftimestamp_units\x18\x08 \x01(\x0e\x32/.tecton_proto.online_store_writer.TimestampUnitR\x0etimestampUnitsB\x0c\n\nstage_typeJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03\"\xc3\x01\n\x13StatusUpdateRequest\x12=\n\x0b\x61nchor_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\nanchorTime\x12`\n\x1ematerialized_raw_data_end_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\x1amaterializedRawDataEndTimeB\x0b\n\ttime_type\"\x90\x01\n\x0f\x44\x65letionRequest\x12\x33\n\x15online_join_keys_path\x18\x01 \x01(\tH\x00R\x12onlineJoinKeysPath\x12<\n\x1aonline_join_keys_full_path\x18\x02 \x01(\tH\x00R\x16onlineJoinKeysFullPathB\n\n\x08location*8\n\rTimestampUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06MILLIS\x10\x01\x12\n\n\x06MICROS\x10\x02\x42 \n\x1c\x63om.tecton.onlinestorewriterP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.online_store_writer.copier__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.tecton.onlinestorewriterP\001'
  _TIMESTAMPUNIT._serialized_start=2526
  _TIMESTAMPUNIT._serialized_end=2582
  _ONLINESTORECOPIERREQUEST._serialized_start=324
  _ONLINESTORECOPIERREQUEST._serialized_end=1297
  _S3STAGE._serialized_start=1299
  _S3STAGE._serialized_end=1350
  _GCSSTAGE._serialized_start=1352
  _GCSSTAGE._serialized_end=1406
  _SNOWFLAKEINTERNALSTAGE._serialized_start=1409
  _SNOWFLAKEINTERNALSTAGE._serialized_end=1607
  _LOCALFILESTAGE._serialized_start=1609
  _LOCALFILESTAGE._serialized_end=1653
  _OBJECTCOPYREQUEST._serialized_start=1656
  _OBJECTCOPYREQUEST._serialized_end=2179
  _STATUSUPDATEREQUEST._serialized_start=2182
  _STATUSUPDATEREQUEST._serialized_end=2377
  _DELETIONREQUEST._serialized_start=2380
  _DELETIONREQUEST._serialized_end=2524
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/diff_options__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/args/diff_options__client.proto\x12\x11tecton_proto.args\x1a google/protobuf/descriptor.proto\"]\n\x11\x46ieldRenameConfig\x12\x1f\n\x0b\x66ormer_name\x18\x01 \x01(\tR\nformerName\x12\'\n\x0f\x63utover_version\x18\x02 \x01(\tR\x0e\x63utoverVersion\"\xc9\x02\n\x0b\x44iffOptions\x12\x39\n\x06update\x18\x01 \x01(\x0e\x32!.tecton_proto.args.UpdateStrategyR\x06update\x12\x1b\n\thide_path\x18\x02 \x01(\x08R\x08hidePath\x12R\n\x0erendering_type\x18\x03 \x01(\x0e\x32+.tecton_proto.args.FcoPropertyRenderingTypeR\rrenderingType\x12P\n\x11\x63ustom_comparator\x18\x04 \x01(\x0e\x32#.tecton_proto.args.CustomComparatorR\x10\x63ustomComparator\x12<\n\x06rename\x18\x05 \x01(\x0b\x32$.tecton_proto.args.FieldRenameConfigR\x06rename*\x85\x02\n\x0eUpdateStrategy\x12\x0c\n\x08RECREATE\x10\x00\x12\x0b\n\x07INPLACE\x10\x01\x12\x12\n\x0eINPLACE_ON_ADD\x10\x03\x12\x1a\n\x16ONE_WAY_INPLACE_ON_ADD\x10\t\x12\x15\n\x11INPLACE_ON_REMOVE\x10\x04\x12\x0b\n\x07PASSIVE\x10\x05\x12\x1e\n\x1aRECREATE_UNLESS_SUPPRESSED\x10\x06\x12\x35\n1RECREATE_UNLESS_SUPPRESSED_INVALIDATE_CHECKPOINTS\x10\x07\x12-\n)RECREATE_UNLESS_SUPPRESSED_RESTART_STREAM\x10\x08*\xc1\x02\n\x18\x46\x63oPropertyRenderingType\x12+\n\'FCO_PROPERTY_RENDERING_TYPE_UNSPECIFIED\x10\x00\x12*\n&FCO_PROPERTY_RENDERING_TYPE_PLAIN_TEXT\x10\x01\x12&\n\"FCO_PROPERTY_RENDERING_TYPE_PYTHON\x10\x02\x12#\n\x1f\x46\x43O_PROPERTY_RENDERING_TYPE_SQL\x10\x03\x12-\n)FCO_PROPERTY_RENDERING_TYPE_ONLY_DECLARED\x10\x04\x12&\n\"FCO_PROPERTY_RENDERING_TYPE_HIDDEN\x10\x05\x12(\n$FCO_PROPERTY_RENDERING_TYPE_REDACTED\x10\x06*\xb2\x05\n\x10\x43ustomComparator\x12\x1b\n\x17\x43USTOM_COMPARATOR_UNSET\x10\x00\x12&\n\"CUSTOM_COMPARATOR_AGGREGATION_NAME\x10\x01\x12\x31\n-CUSTOM_COMPARATOR_OPTION_VALUE_WITH_REDACTION\x10\x02\x12$\n CUSTOM_COMPARATOR_DISPLAY_NOTSET\x10\x03\x12$\n CUSTOM_COMPARATOR_REQUEST_SOURCE\x10\x04\x12,\n(CUSTOM_COMPARATOR_STREAM_PROCESSING_MODE\x10\x06\x12\x34\n0CUSTOM_COMPARATOR_DIFF_OVERRIDE_JOIN_KEYS_AS_MAP\x10\x07\x12$\n CUSTOM_COMPARATOR_BATCH_SCHEDULE\x10\x08\x12(\n$CUSTOM_COMPARATOR_TIME_WINDOW_LEGACY\x10\t\x12!\n\x1d\x43USTOM_COMPARATOR_TIME_WINDOW\x10\n\x12,\n(CUSTOM_COMPARATOR_ONLINE_OFFLINE_ENABLED\x10\x0b\x12#\n\x1f\x43USTOM_COMPARATOR_OFFLINE_STORE\x10\x0c\x12*\n&CUSTOM_COMPARATOR_OFFLINE_STORE_LEGACY\x10\r\x12/\n+CUSTOM_COMPARATOR_UNITY_CATALOG_ACCESS_MODE\x10\x0e\x12&\n\"CUSTOM_COMPARATOR_ENTITY_JOIN_KEYS\x10\x0f\x12+\n\'CUSTOM_COMPARATOR_FEATURE_PARAM_UPGRADE\x10\x10:a\n\x0c\x64iff_options\x12\x1d.google.protobuf.FieldOptions\x18\xd2\t \x01(\x0b\x32\x1e.tecton_proto.args.DiffOptionsR\x0b\x64iffOptionsB\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.diff_options__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(diff_options)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _UPDATESTRATEGY._serialized_start=529
  _UPDATESTRATEGY._serialized_end=790
  _FCOPROPERTYRENDERINGTYPE._serialized_start=793
  _FCOPROPERTYRENDERINGTYPE._serialized_end=1114
  _CUSTOMCOMPARATOR._serialized_start=1117
  _CUSTOMCOMPARATOR._serialized_end=1807
  _FIELDRENAMECONFIG._serialized_start=101
  _FIELDRENAMECONFIG._serialized_end=194
  _DIFFOPTIONS._serialized_start=197
  _DIFFOPTIONS._serialized_end=526
# @@protoc_insertion_point(module_scope)



"""The specs module contains "Tecton object specs", i.e. Python data models for Tecton objects.

Specs provide a unified, frozen (i.e. immutable), and more useful abstraction over args and data protos for use within
the Python SDK.

See the RFC;
https://www.notion.so/tecton/RFC-Unified-SDK-for-Notebook-Driven-Development-a377af9d320f46488ea238e51e2ce656
"""

from tecton_core.specs.data_source_spec import *
from tecton_core.specs.entity_spec import *
from tecton_core.specs.feature_service_spec import *
from tecton_core.specs.feature_view_spec import *
from tecton_core.specs.tecton_object_spec import *
from tecton_core.specs.time_window_spec import *
from tecton_core.specs.time_window_spec import TimeWindowSeriesSpec
from tecton_core.specs.transformation_spec import *



from datetime import timedelta
from functools import reduce
from typing import Dict
from typing import List
from typing import Optional

import attrs
import pendulum
import pyspark
import pyspark.sql.window as spark_window
from pyspark.sql import functions as F
from pyspark.sql.functions import expr
from pyspark.sql.types import ArrayType
from pyspark.sql.types import LongType

from tecton_core import time_utils as core_time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query_consts import TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import valid_from
from tecton_core.query_consts import valid_to
from tecton_core.specs import MaterializedFeatureViewType
from tecton_proto.common.aggregation_function__client_pb2 import AggregationFunction
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2
from tecton_spark.query.node import SparkExecNode
from tecton_spark.time_utils import convert_epoch_to_timestamp_column
from tecton_spark.time_utils import convert_timestamp_to_epoch


@attrs.frozen
class AddAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_format_version: int
    batch_schedule: int
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(F.col(self.timestamp_field), self.feature_store_format_version)
        df = input_df.withColumn(
            anchor_time(),
            anchor_time_val - anchor_time_val % self.batch_schedule,
        )
        return df


@attrs.frozen
class AddRetrievalAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    name: str
    feature_store_format_version: int
    batch_schedule: int
    tile_interval: int
    timestamp_field: str
    is_stream: bool
    data_delay_seconds: Optional[int] = 0

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(
            F.col(self.timestamp_field) - expr(f"interval {self.data_delay_seconds} seconds"),
            self.feature_store_format_version,
        )
        # tile_interval will be 0 for continuous
        if self.tile_interval == 0:
            df = input_df.withColumn(anchor_time(), anchor_time_val)
        else:
            # For stream, we use the tile interval for bucketing since the data is available as soon as
            # the aggregation interval ends.
            # For BAFV, we use the batch schedule to get the last tile written.
            if self.is_stream:
                df = input_df.withColumn(
                    anchor_time(),
                    anchor_time_val - anchor_time_val % self.tile_interval - self.tile_interval,
                )
            else:
                df = input_df.withColumn(
                    anchor_time(),
                    anchor_time_val - anchor_time_val % self.batch_schedule - self.tile_interval,
                )
        return df


@attrs.frozen
class RenameColsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    mapping: Optional[Dict[str, str]]
    drop: Optional[List[str]]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        if self.mapping:
            for old_name, new_name in self.mapping.items():
                if new_name:
                    input_df = input_df.withColumnRenamed(old_name, new_name)
        if self.drop:
            for col in self.drop:
                input_df = input_df.drop(col)
        return input_df


@attrs.frozen
class ConvertEpochToTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_formats: Dict[str, int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        for name, feature_store_format_version in self.feature_store_formats.items():
            input_df = input_df.withColumn(
                name,
                convert_epoch_to_timestamp_column(F.col(name), feature_store_format_version),
            )
        return input_df


@attrs.frozen
class ConvertTimestampToUTCSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_key: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.input_node.to_dataframe(spark)


@attrs.frozen
class AddAnchorTimeColumnsForSawtoothIntervalsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    anchor_time_column_map: Dict[str, timedelta]
    data_delay_seconds: int
    feature_store_format_version: int
    aggregation_tile_interval: int

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        if self.timestamp_field != anchor_time():
            anchor_time_val = convert_timestamp_to_epoch(
                F.col(self.timestamp_field) - expr(f"interval {self.data_delay_seconds} seconds"),
                self.feature_store_format_version,
            )
        else:
            anchor_time_val = F.col(self.timestamp_field)

        columns = [F.col(column) for column in self.input_node.columns]
        new_timestamp_columns = []
        for column_name, unit in self.anchor_time_column_map.items():
            interval_seconds = core_time_utils.convert_timedelta_for_version(unit, self.feature_store_format_version)
            if interval_seconds == 0:
                spark_column = anchor_time_val.alias(column_name)
            else:
                spark_column = (
                    anchor_time_val - anchor_time_val % interval_seconds - self.aggregation_tile_interval
                ).alias(column_name)
            new_timestamp_columns.append(spark_column)
        return input_df.select(*columns, *new_timestamp_columns)


@attrs.frozen
class AdjustAnchorTimeToWindowEndSparkNode(SparkExecNode):
    input_node: SparkExecNode
    anchor_time_columns: List[str]
    aggregation_tile_interval: int

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        output_df = self.input_node.to_dataframe(spark)
        for column in self.anchor_time_columns:
            output_df = output_df.withColumn(column, F.col(column) + self.aggregation_tile_interval)
        return output_df


@attrs.frozen
class AddEffectiveTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    effective_timestamp_name: str
    batch_schedule_seconds: int
    is_stream: bool
    data_delay_seconds: int
    is_temporal_aggregate: bool

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        # batch_schedule = 0 implies feature table.
        if self.batch_schedule_seconds == 0 or self.is_stream:
            effective_timestamp = F.col(self.timestamp_field)
        else:
            slide_str = f"{self.batch_schedule_seconds} seconds"
            timestamp_col = F.col(self.timestamp_field)
            # Timestamp of temporal aggregate is end of the anchor time window. Subtract 1 micro
            # to get the correct bucket for batch schedule.
            if self.is_temporal_aggregate:
                timestamp_col -= expr("interval 1 microseconds")
            window_spec = F.window(timestamp_col, slide_str, slide_str)
            effective_timestamp = window_spec.end + expr(f"interval {self.data_delay_seconds} seconds")

        df = input_df.withColumn(self.effective_timestamp_name, effective_timestamp)
        return df


@attrs.frozen
class AddDurationSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    duration: pendulum.Duration
    new_column_name: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.withColumn(
            self.new_column_name,
            F.col(self.timestamp_field) + expr(f"interval {self.duration.total_seconds()} seconds"),
        )


@attrs.frozen
class SelectDistinctSparkNode(SparkExecNode):
    input_node: SparkExecNode
    columns: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.select(self.columns).distinct()


@attrs.frozen
class AddBooleanPartitionColumnsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    column_to_bool_map: Dict[str, bool]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        for column_name, value in self.column_to_bool_map.items():
            input_df = input_df.withColumn(column_name, F.lit(value))
        return input_df


@attrs.frozen
class ExplodeEventsAndSelectDistinctSparkNode(SparkExecNode):
    """
    Explode the events in the input node and select distinct columns.
    """

    input_node: SparkExecNode
    explode_columns: List[str]
    explode_columns_to_boolean_columns: Dict[str, str]

    def _add_identifier_columns(self, df: pyspark.sql.DataFrame, explode_column: str) -> pyspark.sql.DataFrame:
        columns_to_add = []
        for column in self.explode_columns:
            identifier = self.explode_columns_to_boolean_columns[column]
            if column == explode_column:
                columns_to_add.append(F.lit(True).alias(identifier))
            else:
                columns_to_add.append(F.lit(False).alias(identifier))
        return df.select("*", *columns_to_add)

    def _select_distinct_for_explode_column(
        self, input_df: pyspark.sql.DataFrame, explode_column: str
    ) -> pyspark.sql.DataFrame:
        new_df = self._add_identifier_columns(input_df, explode_column)
        other_columns = [col for col in self.explode_columns if col != explode_column]
        for column in other_columns:
            new_df = new_df.withColumn(column, F.lit(None))
        return new_df.select(*self.columns).distinct()

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        union_df = self._select_distinct_for_explode_column(input_df, self.explode_columns[0])
        if len(self.explode_columns) == 1:
            return union_df

        for explode_column in self.explode_columns[1:]:
            intermediate_df = self._select_distinct_for_explode_column(input_df, explode_column)
            union_df = union_df.union(intermediate_df)
        return union_df


@attrs.frozen
class AggregationSecondaryKeyExplodeSparkNode(SparkExecNode):
    """
    This node returns all <entity, aggregation secondary key> pairs from the input node. It's used for retrieving
    secondary key aggregate features without spine.

    This node is similar to AsofSecondaryKeyExplodeSparkNode but doesn't require as-of join. It uses entities and
    anchor time to look back the max aggregation window time to find all secondary key values, and build a spine with
    all of them.
    """

    input_node: SparkExecNode
    join_keys: List[str]
    aggregation_secondary_key: str
    has_lifetime_aggregate: bool
    earliest_anchor_time_from_window_start: Optional[int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        window_start = (
            spark_window.Window.unboundedPreceding
            if self.has_lifetime_aggregate
            else self.earliest_anchor_time_from_window_start
        )

        window_spec = (
            spark_window.Window.partitionBy(self.join_keys)
            .orderBy([anchor_time()])
            .rangeBetween(window_start, spark_window.Window.currentRow)
        )

        # Collect all secondary key values within the aggregation window, and explode them to build a spine.
        # Note the window based `collect_set` applies to every single row, so if a single join key has multiple
        # aggregation secondary keys with the same anchor time, the result dataframe will have multiple same rows for
        # that particular join key and anchor time. We select distinct here to deduplicate.
        input_df = (
            input_df.withColumn(
                TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL,
                F.collect_set(self.aggregation_secondary_key).over(window_spec),
            )
            .drop(self.aggregation_secondary_key)
            .distinct()
        )
        return input_df.withColumn(
            self.aggregation_secondary_key, F.explode(TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL)
        ).drop(TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL)


@attrs.frozen
class AddUniqueIdSparkNode(SparkExecNode):
    """
    Add a '_tecton_unique_id' column to the input node. This column is used to uniquely identify rows in the input node.

    Warning: The generated unique ID is non-deterministic on Spark, so this column may not be safe to join on.
    """

    input_node: SparkExecNode
    column_name: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.input_node.to_dataframe(spark).withColumn(self.column_name, F.monotonically_increasing_id())


def _is_count_feature(feature: feature_view__data_pb2.AggregateFeature) -> bool:
    return (
        feature.function == AggregationFunction.AGGREGATION_FUNCTION_COUNT
        or feature.function == AggregationFunction.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT
    )


@attrs.frozen
class DeriveValidityPeriodSparkNode(SparkExecNode):
    """
    Derives the `valid_from` and `valid_to` columns from a fully aggregated data frame
    and removes duplicates and rows with default values for all aggregation columns.
    """

    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        join_keys = self.fdw.join_keys
        df = self.input_node.to_dataframe(spark)
        features = self.fdw.fv_spec.features
        window = spark_window.Window.partitionBy(*join_keys).orderBy(self.timestamp_field)

        df = df.withColumn(valid_from(), F.col(self.timestamp_field)).withColumn(
            valid_to(), F.lead(self.timestamp_field).over(window)
        )

        feature_value_update_conditions = [
            ~F.col(column).eqNullSafe(F.col(f"_tecton_prev_{column}")) for column in features
        ]
        track_changes_cols = list(features)

        if self.fdw.serving_ttl is not None:
            df = self._expire_ttl(df)

            # if `valid_from` is not equal to the previous `valid_to` (in the case of TTL expiry),
            # we should not merge the rows even though the feature values are equal
            track_changes_cols.append(valid_to())
            feature_value_update_conditions.append(~F.col(valid_from()).eqNullSafe(F.col(f"_tecton_prev_{valid_to()}")))

        previous_value_columns = [
            F.lag(column).over(window).alias(f"_tecton_prev_{column}") for column in track_changes_cols
        ]
        df = df.select("*", *previous_value_columns)

        # Check if any feature column changed in value, set is_new=1 for rows with changes
        df = df.withColumn(
            "_tecton_is_new", F.when(reduce(lambda x, y: x | y, feature_value_update_conditions), 1).otherwise(0)
        )

        # Cumulatively sum the `is_new` column to create "groups".
        # A group is a set of consecutive rows that has the same aggregated value.
        df = df.withColumn("_tecton_group", F.sum("_tecton_is_new").over(window))

        # if valid_to for a group is "null", we've reached the last event and the "max" should be null
        # since we don't know how long the value is valid
        null_check = F.max(F.when(F.col(valid_to()).isNull(), True).otherwise(False))
        df = df.groupBy(*join_keys, *features, "_tecton_group").agg(
            F.min(valid_from()).alias(valid_from()),
            F.when(null_check, F.lit(None)).otherwise(F.max(valid_to())).alias(valid_to()),
        )

        if self.fdw.fv_spec.type == MaterializedFeatureViewType.TEMPORAL_AGGREGATE:
            df = self._remove_default_values(df)

        return df.drop("_tecton_group", self.timestamp_field)

    def _remove_default_values(self, df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        """
        Removes rows with default values for all aggregation columns.
        """
        aggregate_features = self.fdw.fv_spec.aggregate_features
        is_default_value_conditions = []
        if self.fdw.aggregation_secondary_key:
            # For secondary key aggs, we remove rows that have empty rollup output lists
            column_names = [rollup_output.name for rollup_output in self.fdw.fv_spec.secondary_key_rollup_outputs]
            is_default_value_conditions.extend([F.size(F.col(column)) == 0 for column in column_names])
        else:
            for feature in aggregate_features:
                column_name = feature.output_feature_name
                if isinstance(df.schema[column_name].dataType, ArrayType):
                    all_counts_empty = F.size(F.array_except(column_name, F.array(F.lit(0)))) == 0
                    all_nulls = F.size(F.array_except(column_name, F.array(F.lit(None)))) == 0
                    if _is_count_feature(feature):
                        condition = all_counts_empty | all_nulls
                    else:
                        condition = all_nulls
                else:
                    if _is_count_feature(feature):
                        condition = F.col(column_name).isNull() | (F.col(column_name) == 0)
                    else:
                        condition = F.col(column_name).isNull()
                is_default_value_conditions.append(condition)

        return df.filter(~reduce(lambda x, y: x & y, is_default_value_conditions))

    def _expire_ttl(self, df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        """
        Trims `valid_to` values according to TTL for non-aggregate feature views
        """
        ttl_seconds = self.fdw.serving_ttl.total_seconds()
        should_trim_valid_to = F.col(valid_to()).isNull() | (
            (F.col(valid_to()).cast(LongType()) - F.col(valid_from()).cast(LongType())) > ttl_seconds
        )
        df = df.withColumn(
            valid_to(),
            F.when(should_trim_valid_to, F.col(valid_from()) + expr(f"INTERVAL {ttl_seconds} SECONDS")).otherwise(
                F.col(valid_to())
            ),
        )

        return df



import codecs
import os.path
import pathlib

import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    path = pathlib.Path(rel_path).resolve()
    if path.exists():
        for line in read(rel_path).splitlines():
            if line.startswith("VERSION"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        exc = RuntimeError("Unable to find version string.")
        raise exc
    else:
        # NOTE: we use `99.99.99.dev` here instead of raising an exception, since
        # pip-tools executes this file as part of dependency resolution. And our
        # version file isn't available nor necessary for it.
        return "99.99.99.dev"


setuptools.setup(version=get_version("tecton/_gen_version.py"))



import contextlib
import json
import logging
import os
import pathlib
import sys
import threading
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

import boto3

from tecton_core import errors


logging.getLogger("boto3").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)

_ConfigSettings = Dict[str, Any]

_CONFIG_OVERRIDES: _ConfigSettings = {}


class ConfSource(Enum):
    # (Always supported) This key can be overriden in the Python runtime.
    SESSION_OVERRIDE = 1
    # (Always supported) This key can be read from the local environment.
    OS_ENV = 2
    # This key can be written to and read from the tecton config file.
    LOCAL_TECTON_CONFIG = 3
    # This key can be written to and read from the tecton tokens file.
    LOCAL_TECTON_TOKENS = 4
    # This key can read from the MDS get-configs endpoint. Note that this will not return anything until after the
    # first MDS RPC is made, which is racy because e.g. @sdk_public_method makes asynchronous MDS calls.
    #
    # BLOCKING_REMOTE_MDS_CONFIG is not racy, but can't be used everywhere without causing MDS requests to happen during
    # initialization. We can clean up this whole mess when confs from secret managers are explicit and AWS_REGION is no
    # longer required.
    REMOTE_MDS_CONFIG = 5
    # Like REMOTE_MDS_CONFIG, but blocks until the initial configuration RPC from MDS is complete.
    BLOCKING_REMOTE_MDS_CONFIG = 6
    # This key can read from the Databricks secrets manager.
    DATABRICKS_SECRET = 7
    # This key can read from the AWS secrets manager.
    AWS_SECRET_MANAGER = 8
    # The default value of this key was used. Used for debug printing. Not an "allowable" source.
    DEFAULT = 9
    # A value for this key was not found. Used for debug printing. Not an "allowable" source.
    NOT_FOUND = 10


RUNTIME_ALLOWED_SOURCES = (ConfSource.SESSION_OVERRIDE, ConfSource.OS_ENV)
DEFAULT_ALLOWED_SOURCES = (
    ConfSource.SESSION_OVERRIDE,
    ConfSource.OS_ENV,
    ConfSource.REMOTE_MDS_CONFIG,
    ConfSource.DATABRICKS_SECRET,
    ConfSource.AWS_SECRET_MANAGER,
)


class _Debugger(object):
    @classmethod
    def _debug_enabled(cls) -> bool:
        debug_value = _get_runtime_only("TECTON_DEBUG")
        return debug_value is not None and debug_value.lower() in ("1", "true", "yes")

    @classmethod
    def preamble(cls, key: str) -> None:
        if cls._debug_enabled():
            print(f"Looking up {key}", file=sys.stderr)

    @classmethod
    def print(cls, src: ConfSource, key: str, val: Optional[str] = None, details: Optional[str] = None) -> None:
        if not cls._debug_enabled():
            return

        if src == ConfSource.NOT_FOUND:
            print(f"Unable to find {key}\n", file=sys.stderr)
            return

        details_str = f"({details})" if details else ""
        val_str = val if val is not None else "not found"
        symbol_str = "[x]" if val is not None else "[ ]"
        print(symbol_str, f"{key} in {src.name} -> {val_str}", details_str, file=sys.stderr)


def set(key: str, value: str) -> None:
    _set(key, value)


def unset(key: str) -> None:
    del _CONFIG_OVERRIDES[key]


def _set(key: str, value: str) -> None:
    _CONFIG_OVERRIDES[key] = value


@contextlib.contextmanager
def _temporary_set(key: str, value: str) -> Iterator[None]:
    curr_val = _CONFIG_OVERRIDES.get(key)
    _CONFIG_OVERRIDES[key] = value
    try:
        yield
    finally:
        if curr_val:
            _CONFIG_OVERRIDES[key] = curr_val
        else:
            del _CONFIG_OVERRIDES[key]


def _does_key_have_valid_prefix(key: str) -> bool:
    for prefix in _VALID_KEY_PREFIXES:
        if key.startswith(prefix):
            return True
    return False


def _get(key: str) -> Optional[str]:
    """Get the config value for the given key, or return None if not found."""
    _Debugger.preamble(key)

    if key in _VALID_KEYS_TO_ALLOWED_SOURCES:
        allowed_sources = _VALID_KEYS_TO_ALLOWED_SOURCES[key]
    elif _does_key_have_valid_prefix(key):
        allowed_sources = DEFAULT_ALLOWED_SOURCES
    else:
        msg = f"Tried accessing invalid configuration key '{key}'"
        raise errors.TectonInternalError(msg)

    # Session-scoped override.
    if ConfSource.SESSION_OVERRIDE in allowed_sources:
        val = _CONFIG_OVERRIDES.get(key)
        _Debugger.print(ConfSource.SESSION_OVERRIDE, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # Environment variable.
    if ConfSource.OS_ENV in allowed_sources:
        val = os.environ.get(key)
        _Debugger.print(ConfSource.OS_ENV, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # ~/.tecton/config
    if ConfSource.LOCAL_TECTON_CONFIG in allowed_sources:
        val = _LOCAL_TECTON_CONFIG.get(key)
        _Debugger.print(ConfSource.LOCAL_TECTON_CONFIG, key, val, details=str(_LOCAL_TECTON_CONFIG_FILE))
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # ~/.tecton/config.tokens
    if ConfSource.LOCAL_TECTON_TOKENS in allowed_sources:
        val = _LOCAL_TECTON_TOKENS.get(key)
        _Debugger.print(ConfSource.LOCAL_TECTON_TOKENS, key, val, details=str(_LOCAL_TECTON_TOKENS_FILE))
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    if ConfSource.BLOCKING_REMOTE_MDS_CONFIG in allowed_sources:
        _force_initialize_mds_config()

    # Config from MDS
    if ConfSource.REMOTE_MDS_CONFIG in allowed_sources or ConfSource.BLOCKING_REMOTE_MDS_CONFIG in allowed_sources:
        with _mds_config_lock:
            val = _remote_mds_configs.get(key)
        _Debugger.print(ConfSource.REMOTE_MDS_CONFIG, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    if _get_runtime_env() == TectonEnv.UNKNOWN:
        # Fallback attempt to set env if user has not set it.
        _set_tecton_runtime_env()

    # NOTE: although originally intended for internal use. Some customers have
    # found this configuration or have required this behavior. Care should be
    # exercised in changing any behavior here to ensure no customer breakages.
    if ConfSource.DATABRICKS_SECRET in allowed_sources and not _get_runtime_only("TECTON_CONF_DISABLE_DBUTILS"):
        # Databricks secrets
        for scope in _get_secret_scopes():
            value = _get_from_db_secrets(key, scope)
            _Debugger.print(ConfSource.DATABRICKS_SECRET, key, value, details=f"{scope}:{key}")
            if value is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
                return value

    # NOTE: although originally intended for internal use. Some customers have
    # found this configuration or have required this behavior. Care should be
    # exercised in changing any behavior here to ensure no customer breakages.
    if ConfSource.AWS_SECRET_MANAGER in allowed_sources and not _get_runtime_only("TECTON_CONF_DISABLE_AWS_SECRETS"):
        # AWS secret manager
        for scope in _get_secret_scopes():
            value = _get_from_secretsmanager(key, scope)
            if value is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
                return value

    if key in _DEFAULTS:
        value = _DEFAULTS[key]()
        _Debugger.print(ConfSource.DEFAULT, key, value)
        return value

    _Debugger.print(ConfSource.NOT_FOUND, key)
    return None


def _get_runtime_only(key: str) -> Optional[str]:
    """An alternate _get() that will look up only from runtime sources. Used to avoid infinite loops or network calls"""
    if key not in _VALID_KEYS_TO_ALLOWED_SOURCES:
        msg = f"_get_runtime_only should only used with valid keys. {key}"
        raise errors.TectonInternalError(msg)

    # Session-scoped override.
    val = _CONFIG_OVERRIDES.get(key)
    if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
        return val

    # Environment variable.
    val = os.environ.get(key)
    if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
        return val

    if key in _DEFAULTS:
        value = _DEFAULTS[key]()
        return value

    return None


def get_or_none(key: str) -> Optional[str]:
    return _get(key)


def get_or_raise(key: str) -> str:
    val = _get(key)
    if val is None:
        msg = f"{key} not set"
        raise errors.TectonInternalError(msg)
    return val


def get_bool(key: str) -> bool:
    val = _get(key)
    if val is None:
        return False
    # bit of a hack for if people set a boolean value in a local override
    if isinstance(val, bool):
        return val
    if not isinstance(val, str):
        msg = f"{key} should be an instance of str, not {type(val)}"
        raise ValueError(msg)
    if val.lower() in {"yes", "true"}:
        return True
    if val.lower() in {"no", "false"}:
        return False
    msg = f"{key} should be 'true' or 'false', not {val}"
    raise ValueError(msg)


def _model_cache_directory() -> str:
    base_cache_dir = os.environ.get("XDG_CACHE_HOME", str(pathlib.Path.home() / ".cache"))
    return os.path.join(base_cache_dir, "tecton", "models")


# Internal

_LOCAL_TECTON_CONFIG_FILE = Path(os.environ.get("TECTON_CONFIG_PATH", Path.home() / ".tecton/config"))
_LOCAL_TECTON_TOKENS_FILE = _LOCAL_TECTON_CONFIG_FILE.with_suffix(".tokens")

_VALID_KEYS_TO_ALLOWED_SOURCES = {
    "API_SERVICE": (*DEFAULT_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_CONFIG),
    "FEATURE_SERVICE": (*DEFAULT_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_CONFIG),
    "INGESTION_SERVICE": (*DEFAULT_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_CONFIG),
    "CLI_CLIENT_ID": (*DEFAULT_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_CONFIG),
    "TECTON_WORKSPACE": (*DEFAULT_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_CONFIG),
    "TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.BLOCKING_REMOTE_MDS_CONFIG,
    ),
    "TECTON_BATCH_COMPUTE_MODE": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.BLOCKING_REMOTE_MDS_CONFIG,
    ),
    "OAUTH_ACCESS_TOKEN": (*RUNTIME_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_TOKENS),
    "OAUTH_ACCESS_TOKEN_EXPIRATION": (*RUNTIME_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_TOKENS),
    "OAUTH_REFRESH_TOKEN": (*RUNTIME_ALLOWED_SOURCES, ConfSource.LOCAL_TECTON_TOKENS),
    # TECTON_CLUSTER_NAME is needed for looking up AWS and Databricks secrets. CLUSTER_REGION is used for looking up AWS
    # secrets. To avoid an infinite loop, these keys cannot be looked up from those sources.
    "TECTON_CLUSTER_NAME": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.REMOTE_MDS_CONFIG,
    ),
    "CLUSTER_REGION": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.REMOTE_MDS_CONFIG,
        ConfSource.DATABRICKS_SECRET,
    ),
    # NOTE: TECTON_CONF* are meant for Tecton internal use.
    # TODO(TEC-8744): improve tecton.conf configurations such that end users have more
    # control of where secrets are fetched from.
    "SPARK_REDSHIFT_TEMP_DIR": RUNTIME_ALLOWED_SOURCES,
    "TECTON_CONF_DISABLE_DBUTILS": RUNTIME_ALLOWED_SOURCES,
    "TECTON_CONF_DISABLE_AWS_SECRETS": RUNTIME_ALLOWED_SOURCES,
    "TECTON_DEBUG": RUNTIME_ALLOWED_SOURCES,
    "TECTON_RUNTIME_ENV": RUNTIME_ALLOWED_SOURCES,
    "TECTON_RUNTIME_MODE": RUNTIME_ALLOWED_SOURCES,
    "TECTON_SKIP_OBJECT_VALIDATION": RUNTIME_ALLOWED_SOURCES,
    "TECTON_REQUIRE_SCHEMA": RUNTIME_ALLOWED_SOURCES,
    # TECTON_FORCE_FUNCTION_SERIALIZATION is a parameter used by some users in unit testing. Replacing or making changes
    # to the semantics of this flag will require upgrade/migration guidance. However, this is not a "public/stable"
    # flag, so feel free to make changes as long as there is an upgrade path.
    "TECTON_FORCE_FUNCTION_SERIALIZATION": RUNTIME_ALLOWED_SOURCES,
    "TECTON_REPO_IGNORE_ALL_HIDDEN_DIRS": RUNTIME_ALLOWED_SOURCES,
    "SKIP_OBJECT_VERSION_CHECK": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_DEBUG": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_BENCHMARK": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_PERSIST_DB": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_MEMORY_LIMIT": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_NTHREADS": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_EXTENSION_REPO": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_ALLOW_CACHE_EXTENSION": RUNTIME_ALLOWED_SOURCES,
    "ATHENA_S3_PATH": DEFAULT_ALLOWED_SOURCES,
    "ATHENA_DATABASE": DEFAULT_ALLOWED_SOURCES,
    "ENABLE_TEMPO": DEFAULT_ALLOWED_SOURCES,
    "QUERY_REWRITE_ENABLED": DEFAULT_ALLOWED_SOURCES,
    "AWS_ACCESS_KEY_ID": DEFAULT_ALLOWED_SOURCES,
    "AWS_SECRET_ACCESS_KEY": DEFAULT_ALLOWED_SOURCES,
    "AWS_SESSION_TOKEN": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_HOST": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_PORT": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_USERNAME": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_DATABASE": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SPARK_DRIVER_LOCAL_IP": DEFAULT_ALLOWED_SOURCES,
    "METADATA_SERVICE": DEFAULT_ALLOWED_SOURCES,
    "TECTON_API_KEY": DEFAULT_ALLOWED_SOURCES,
    "QUERYTREE_SHORT_SQL_ENABLED": RUNTIME_ALLOWED_SOURCES,
    "QUERYTREE_SHORT_SQL_TYPE": RUNTIME_ALLOWED_SOURCES,
    "REDSHIFT_USER": DEFAULT_ALLOWED_SOURCES,
    "REDSHIFT_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SKIP_FEATURE_TIMESTAMP_VALIDATION": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_ACCOUNT_IDENTIFIER": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_DEBUG": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_SHORT_SQL_ENABLED": DEFAULT_ALLOWED_SOURCES,
    # Whether to break up long SQL statements into temporary views for Snowflake
    "SNOWFLAKE_TEMP_TABLE_ENABLED": DEFAULT_ALLOWED_SOURCES,
    # Whether to break up long SQL statements with temporary tables for Snowflake, takes precedence over SNOWFLAKE_SHORT_SQL_ENABLED
    "SNOWFLAKE_USER": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_WAREHOUSE": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_DATABASE": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_PANDAS_ODFV_ENABLED": RUNTIME_ALLOWED_SOURCES,
    "USE_DEPRECATED_SNOWFLAKE_RETRIEVAL": RUNTIME_ALLOWED_SOURCES,
    "USE_DEPRECATED_ATHENA_RETRIEVAL": RUNTIME_ALLOWED_SOURCES,
    # Instead of the querytree code path, use the old snowflake ghf implementation
    "REDIS_AUTH_TOKEN": DEFAULT_ALLOWED_SOURCES,
    "TECTON_PYTHON_ODFV_OUTPUT_SCHEMA_CHECK_ENABLED": RUNTIME_ALLOWED_SOURCES,
    "USE_LOCAL_OFFLINE_STORE_CREDENTIALS": DEFAULT_ALLOWED_SOURCES,
    "SKIP_REPO_CONFIG_INIT": DEFAULT_ALLOWED_SOURCES,  # Used by itests only
    "ARROW_BATCH_READ_AHEAD": DEFAULT_ALLOWED_SOURCES,
    "ARROW_FRAGMENT_READ_AHEAD": DEFAULT_ALLOWED_SOURCES,
    "DUCKDB_DISK_SPILLING_ENABLED": DEFAULT_ALLOWED_SOURCES,
    "MODEL_CACHE_DIRECTORY": RUNTIME_ALLOWED_SOURCES,
    "DUCKDB_ENABLE_OPTIMIZED_FULL_AGG": DEFAULT_ALLOWED_SOURCES,
    "ALLOW_NULL_FEATURES": DEFAULT_ALLOWED_SOURCES,
    "TECTON_STRIP_TIMEZONE_FROM_FEATURE_VALUES": DEFAULT_ALLOWED_SOURCES,
}

_VALID_KEY_PREFIXES = ["SECRET_"]

_DEFAULTS = {
    "TECTON_WORKSPACE": (lambda: "prod"),
    "FEATURE_SERVICE": (lambda: _get("API_SERVICE")),
    "ENABLE_TEMPO": (lambda: "false"),
    "QUERY_REWRITE_ENABLED": (lambda: "true"),
    "TECTON_REPO_IGNORE_ALL_HIDDEN_DIRS": (lambda: "true"),
    "QUERYTREE_SHORT_SQL_TYPE": (lambda: "table"),
    "DUCKDB_EXTENSION_REPO": (
        lambda: "http://s3.us-west-2.amazonaws.com/tecton.ai.public/duckdb/tecton-extension/{version}"
    ),
    "USE_DEPRECATED_SNOWFLAKE_RETRIEVAL": (lambda: "false"),
    "TECTON_PYTHON_ODFV_OUTPUT_SCHEMA_CHECK_ENABLED": (lambda: "true"),
    "ARROW_BATCH_READ_AHEAD": (lambda: "64"),
    "ARROW_FRAGMENT_READ_AHEAD": (lambda: "16"),
    "DUCKDB_DISK_SPILLING_ENABLED": (lambda: "true"),
    "MODEL_CACHE_DIRECTORY": (lambda: _model_cache_directory()),
    "DUCKDB_ENABLE_OPTIMIZED_FULL_AGG": (lambda: "true"),
    "TECTON_SKIP_OBJECT_VALIDATION": (lambda: "false"),
    "TECTON_REQUIRE_SCHEMA": (lambda: "true"),
    "ALLOW_NULL_FEATURES": (lambda: "true"),
    # TODO: Consider defaulting to false in SDK 1.0
    "TECTON_STRIP_TIMEZONE_FROM_FEATURE_VALUES": (lambda: "true"),
}

_mds_config_lock = threading.Lock()
_remote_mds_configs: _ConfigSettings = {}

_is_interactive_notebook = None

_is_running_on_databricks_cache = None
_is_running_on_emr_cache = None
TectonEnv = Enum("TectonEnv", "DATABRICKS EMR UNKNOWN")

save_tecton_configs_enabled = False  # only save configs for CLI


def _is_running_on_databricks() -> bool:
    """Whether we're running in Databricks notebook or not."""
    global _is_running_on_databricks_cache
    if _is_running_on_databricks_cache is None:
        main = __import__("__main__")
        filename = os.path.basename(getattr(main, "__file__", ""))
        is_python_shell = filename == "PythonShell.py"
        is_databricks_env = "DBUtils" in main.__dict__
        _is_running_on_databricks_cache = is_python_shell and is_databricks_env
    return _is_running_on_databricks_cache


def _is_running_on_emr() -> bool:
    """Whether we're running in EMR notebook or not."""
    global _is_running_on_emr_cache
    if _is_running_on_emr_cache is None:
        _is_running_on_emr_cache = "EMR_CLUSTER_ID" in os.environ
    return _is_running_on_emr_cache


def _set_tecton_runtime_env():
    key = "TECTON_RUNTIME_ENV"
    if _is_running_on_databricks():
        set(key, "DATABRICKS")
    elif _is_running_on_emr():
        set(key, "EMR")
    else:
        set(key, "UNKNOWN")


def _is_mode_materialization() -> bool:
    runtime_mode = get_or_none("TECTON_RUNTIME_MODE")
    return runtime_mode == "MATERIALIZATION"


def _get_runtime_env() -> TectonEnv:
    # Use _get_runtime_only() and not _get() to avoid an infinite loop.
    runtime_env = _get_runtime_only("TECTON_RUNTIME_ENV")
    if runtime_env == "DATABRICKS":
        return TectonEnv.DATABRICKS
    elif runtime_env == "EMR":
        return TectonEnv.EMR
    else:
        return TectonEnv.UNKNOWN


def _get_dbutils():
    # Returns dbutils import. Only works in Databricks notebook environment
    import IPython

    return IPython.get_ipython().user_ns["dbutils"]


def enable_save_tecton_configs() -> None:
    global save_tecton_configs_enabled
    save_tecton_configs_enabled = True


def _get_keys_with_allowed_source(source: ConfSource) -> List[str]:
    """Returns all the keys that have the allowed ConfSource."""
    return [key for key, allowed_sources in _VALID_KEYS_TO_ALLOWED_SOURCES.items() if source in allowed_sources]


def save_tecton_configs() -> None:
    _save_tecton_config(_LOCAL_TECTON_CONFIG_FILE, _get_keys_with_allowed_source(ConfSource.LOCAL_TECTON_CONFIG))


def _save_tecton_config(path: Path, keys: Iterable[str]) -> None:
    tecton_config = {key: get_or_none(key) for key in keys if get_or_none(key) is not None}
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as f:
        json.dump(tecton_config, f, sort_keys=True, indent=2)
        f.write("\n")


def _delete_tecton_config(path: Path) -> None:
    try:
        os.remove(path)
    except Exception:
        pass


# Get key by looking in TECTON_CLUSTER_NAME'd scope and falling back to "tecton"
def _get_secret_scopes() -> List[str]:
    cluster_name = get_or_none("TECTON_CLUSTER_NAME")
    secret_scopes = []
    if cluster_name:
        secret_prefix = cluster_name if cluster_name.startswith("tecton-") else f"tecton-{cluster_name}"
        secret_scopes.append(secret_prefix)
    secret_scopes.append("tecton")
    return secret_scopes


# Initializing boto3 clients is expensive and not thread-safe (thanks boto3)
# The lru_cache decorator ensures the client is initialized only once and in thread-safe way
@lru_cache()
def _get_secretsmanager_client():
    if _is_mode_materialization():
        region_name = None
    else:
        region_name = get_or_none("CLUSTER_REGION")
    return boto3.client("secretsmanager", region_name=region_name)


def _get_from_secretsmanager(key: str, scope: str) -> Optional[str]:
    try:
        # Try to Grab secret from AWS secrets manager
        aws_secret_client = _get_secretsmanager_client()

        secret = aws_secret_client.get_secret_value(SecretId=f"{scope}/{key}")
        value = secret["SecretString"]
        _Debugger.print(ConfSource.AWS_SECRET_MANAGER, key, value, details=f"{scope}/{key}")
        return value
    except Exception as e:
        # If it's not found, just return None and don't give verbose error details. boto3 is garbage so the only way
        # to detect this is by looking at the type name.
        exception_name = type(e).__name__
        if exception_name == "ResourceNotFoundException":
            _Debugger.print(ConfSource.AWS_SECRET_MANAGER, key, details=f"{scope}/{key}")
        else:
            _Debugger.print(
                ConfSource.AWS_SECRET_MANAGER,
                key,
                details=f"Failed to retrieve secret {scope}/{key}: {exception_name}: {e}",
            )
        return None


def _get_from_db_secrets(key: str, scope: str) -> Optional[str]:
    try:
        dbutils = _get_dbutils()
        return dbutils.secrets.get(scope, key)
    except Exception:
        return None


def set_login_configs(
    cli_client_id: str,
    api_service: str,
    feature_service: str,
) -> None:
    (set("CLI_CLIENT_ID", cli_client_id),)
    set("API_SERVICE", api_service)
    set("FEATURE_SERVICE", feature_service)

    if save_tecton_configs_enabled:
        # For notebook environments, don't save the configs and tokens because notebooks
        # attached to the same cluster share a filesystem.
        save_tecton_configs()


def set_okta_tokens(access_token, access_token_expiration, refresh_token=None):
    set("OAUTH_ACCESS_TOKEN", access_token)
    set("OAUTH_ACCESS_TOKEN_EXPIRATION", access_token_expiration)
    if refresh_token:
        set("OAUTH_REFRESH_TOKEN", refresh_token)

    if save_tecton_configs_enabled:
        # For notebook environments, don't save the configs and tokens because notebooks
        # attached to the same cluster share a filesystem.
        save_okta_tokens()


def save_okta_tokens():
    _save_tecton_config(_LOCAL_TECTON_TOKENS_FILE, _get_keys_with_allowed_source(ConfSource.LOCAL_TECTON_TOKENS))


def delete_okta_tokens():
    for key in (
        "OAUTH_ACCESS_TOKEN",
        "OAUTH_REFRESH_TOKEN",
        "OAUTH_ACCESS_TOKEN_EXPIRATION",
    ):
        try:
            unset(key)
        except KeyError:
            pass
    _delete_tecton_config(_LOCAL_TECTON_TOKENS_FILE)


def _read_json_config(file_path: Path) -> _ConfigSettings:
    """If the file exists, reads it and returns parsed JSON. Otherwise returns empty dictionary."""
    if not file_path.exists():
        return {}
    content = file_path.read_text()
    if not content:
        return {}
    try:
        return json.loads(content)
    except json.decoder.JSONDecodeError as e:
        raise ValueError(
            f"Unable to decode JSON configuration file {file_path} ({str(e)}). "
            + "To regenerate configuration, delete this file and run `tecton login`."
        )


def validate_api_service_url(url: str) -> None:
    """Validate Tecton API URL.
    Returns nothing for valid URLs or raises an error."""
    if "localhost" in url or "ingress" in url:
        return
    if not url.endswith("/api"):
        msg = f'Tecton API URL ("{url}") should be formatted "https://<deployment>.tecton.ai/api"'
        raise errors.TectonAPIValidationError(msg)


def tecton_url() -> str:
    api_service = get_or_none("API_SERVICE")
    if not api_service:
        msg = "Tecton URL not set. Please configure API_SERVICE or use tecton.set_credentials(tecton_url=<url>)."
        raise errors.TectonAPIInaccessibleError(msg)
    validate_api_service_url(api_service)
    url, _, _ = api_service.partition("/api")
    return url


# Config values written to and read from the local .tecton/config file.
_LOCAL_TECTON_CONFIG: _ConfigSettings = {}

# Config values read from the local .tecton/config.tokens file.
_LOCAL_TECTON_TOKENS: _ConfigSettings = {}


def _init_configs():
    if not _is_mode_materialization():
        global _LOCAL_TECTON_CONFIG
        _LOCAL_TECTON_CONFIG = _read_json_config(_LOCAL_TECTON_CONFIG_FILE)

        global _LOCAL_TECTON_TOKENS
        _LOCAL_TECTON_TOKENS = _read_json_config(_LOCAL_TECTON_TOKENS_FILE)


def _init_metadata_server_config(mds_response):
    global _remote_mds_configs
    with _mds_config_lock:
        _remote_mds_configs = dict(mds_response.key_values)


def _force_initialize_mds_config():
    runtime_mode = get_or_none("TECTON_RUNTIME_MODE")
    if runtime_mode in ("MATERIALIZATION", "EVALUATION"):
        return
    from tecton._internals import metadata_service  # noqa: TID251

    metadata_service.instance()


def _clear_metadata_server_config():
    global _remote_mds_configs
    with _mds_config_lock:
        _remote_mds_configs = {}


_init_configs()



"""This files contains utilities for running feature_table.ingest()."""

import io
import logging

import pandas
from pyspark.sql import dataframe as pyspark_dataframe
from pyspark.sql.types import ArrayType
from pyspark.sql.types import DoubleType
from pyspark.sql.types import FloatType
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StructType

from tecton import tecton_context
from tecton_core import errors
from tecton_core import http
from tecton_core import schema
from tecton_core.arrow import PARQUET_WRITE_OPTIONS_KWARGS
from tecton_core.pandas_compat import pandas_to_spark
from tecton_spark import schema_spark_utils


logger = logging.getLogger(__name__)


def upload_df_pandas(upload_url: str, df: pandas.DataFrame, parquet: bool = True):
    out_buffer = io.BytesIO()
    if parquet:
        df.to_parquet(out_buffer, index=False, engine="pyarrow", **PARQUET_WRITE_OPTIONS_KWARGS)
    else:
        df.to_csv(out_buffer, index=False, header=False)

    # Maximum 1GB per ingestion
    if out_buffer.__sizeof__() > 1_000_000_000:
        raise errors.FT_DF_TOO_LARGE

    r = http.session().put(upload_url, data=out_buffer.getvalue())
    if r.status_code != 200:
        raise errors.FT_UPLOAD_FAILED(r.reason)


def convert_pandas_to_spark_df(df: pandas.DataFrame, view_schema: schema.Schema) -> pyspark_dataframe.DataFrame:
    tc = tecton_context.TectonContext.get_instance()
    spark = tc._spark
    spark_df = pandas_to_spark(spark, df)

    converted_schema = _convert_ingest_schema(spark_df.schema, view_schema)

    if converted_schema != spark_df.schema:
        spark_df = pandas_to_spark(df, schema=converted_schema)

    return spark_df


def _convert_ingest_schema(ingest_schema: StructType, view_schema: schema.Schema) -> StructType:
    """Convert pandas-derived spark schema to Tecton-compatible schema for Feature Tables.

    The Pandas to Spark dataframe conversion implicitly derives the Spark schema.
    We handle converting/correcting for some type conversions where the derived schema and the feature table schema do not match.
    """
    ft_columns = schema_spark_utils.column_name_spark_data_types(view_schema)
    ingest_columns = schema_spark_utils.column_name_spark_data_types(
        schema_spark_utils.schema_from_spark(ingest_schema)
    )

    converted_ingest_schema = StructType()
    int_converted_columns = []

    for col_name, col_type in ingest_columns:
        if col_type == LongType() and (col_name, IntegerType()) in ft_columns:
            int_converted_columns.append(col_name)
            converted_ingest_schema.add(col_name, IntegerType())
        elif col_type == ArrayType(DoubleType()) and (col_name, ArrayType(FloatType())) in ft_columns:
            converted_ingest_schema.add(col_name, ArrayType(FloatType()))
        else:
            converted_ingest_schema.add(col_name, col_type)

    if int_converted_columns:
        logger.warning(
            f"Tecton is casting field(s) {', '.join(int_converted_columns)} to type Integer (was type Long). To remove this warning, use a Long type in the schema."
        )

    return converted_ingest_schema



import getpass
from typing import Optional


_current_username = None


def _get_databricks_user() -> Optional[str]:
    try:
        import IPython

        ipy = IPython.get_ipython()
        if ipy is None:
            return None
        dbutils = ipy.user_ns.get("dbutils")
        if dbutils is None:
            return None
        else:
            return dbutils.notebook.entry_point.getDbutils().notebook().getContext().tags().apply("user")
    except ImportError as e:
        pass
    return None


def _get_system_user():
    return getpass.getuser()


def get_current_username() -> str:
    global _current_username
    if _current_username is None:
        _current_username = _get_current_username_uncached()
    return _current_username


def _get_current_username_uncached() -> str:
    for f in (_get_databricks_user, _get_system_user):
        user = f()
        if user:
            return user
    return ""
